import aiorwlock
import asyncio
from importlib import import_module
import inspect
import logging
import pickle
import time
from typing import Any, Callable, Optional, Tuple, Dict

import starlette.responses

import ray
from ray import cloudpickle
from ray.actor import ActorClass, ActorHandle
from ray.remote_function import RemoteFunction
from ray.util import metrics
from ray._private.async_compat import sync_to_async

from ray.serve._private.autoscaling_metrics import start_metrics_pusher
from ray.serve._private.common import HEALTH_CHECK_CONCURRENCY_GROUP, ReplicaTag
from ray.serve.config import DeploymentConfig
from ray.serve._private.constants import (
    HEALTH_CHECK_METHOD,
    RECONFIGURE_METHOD,
    DEFAULT_LATENCY_BUCKET_MS,
    SERVE_LOGGER_NAME,
    SERVE_NAMESPACE,
)
from ray.serve.deployment import Deployment
from ray.serve.exceptions import RayServeException
from ray.serve._private.http_util import ASGIHTTPSender
from ray.serve._private.logging_utils import access_log_msg, configure_component_logger
from ray.serve._private.router import Query, RequestMetadata
from ray.serve._private.utils import (
    parse_import_path,
    parse_request_item,
    wrap_to_ray_error,
    merge_dict,
)
from ray.serve._private.version import DeploymentVersion

logger = logging.getLogger(SERVE_LOGGER_NAME)


def _format_replica_actor_name(deployment_name: str):
    return f"ServeReplica:{deployment_name}"


def create_replica_wrapper(name: str):
    """Creates a replica class wrapping the provided function or class.

    This approach is picked over inheritance to avoid conflict between user
    provided class and the RayServeReplica class.
    """

    # TODO(architkulkarni): Add type hints after upgrading cloudpickle
    class RayServeWrappedReplica(object):
        async def __init__(
            self,
            deployment_name,
            replica_tag,
            serialized_deployment_def: bytes,
            serialized_init_args: bytes,
            serialized_init_kwargs: bytes,
            deployment_config_proto_bytes: bytes,
            version: DeploymentVersion,
            controller_name: str,
            detached: bool,
        ):
            configure_component_logger(
                component_type="deployment",
                component_name=deployment_name,
                component_id=replica_tag,
            )

            deployment_def = cloudpickle.loads(serialized_deployment_def)

            if isinstance(deployment_def, str):
                import_path = deployment_def
                module_name, attr_name = parse_import_path(import_path)
                deployment_def = getattr(import_module(module_name), attr_name)
                # For ray or serve decorated class or function, strip to return
                # original body
                if isinstance(deployment_def, RemoteFunction):
                    deployment_def = deployment_def._function
                elif isinstance(deployment_def, ActorClass):
                    deployment_def = deployment_def.__ray_metadata__.modified_class
                elif isinstance(deployment_def, Deployment):
                    logger.warning(
                        f'The import path "{import_path}" contains a '
                        "decorated Serve deployment. The decorator's settings "
                        "are ignored when deploying via import path."
                    )
                    deployment_def = deployment_def.func_or_class

            init_args = cloudpickle.loads(serialized_init_args)
            init_kwargs = cloudpickle.loads(serialized_init_kwargs)

            deployment_config = DeploymentConfig.from_proto_bytes(
                deployment_config_proto_bytes
            )

            if inspect.isfunction(deployment_def):
                is_function = True
            elif inspect.isclass(deployment_def):
                is_function = False
            else:
                assert False, (
                    "deployment_def must be function, class, or "
                    "corresponding import path. Instead, it's type was "
                    f"{type(deployment_def)}."
                )

            # Set the controller name so that serve.connect() in the user's
            # code will connect to the instance that this deployment is running
            # in.
            ray.serve.context._set_internal_replica_context(
                deployment_name,
                replica_tag,
                controller_name,
                servable_object=None,
            )

            assert controller_name, "Must provide a valid controller_name"

            controller_handle = ray.get_actor(
                controller_name, namespace=SERVE_NAMESPACE
            )

            # This closure initializes user code and finalizes replica
            # startup. By splitting the initialization step like this,
            # we can already access this actor before the user code
            # has finished initializing.
            # The supervising state manager can then wait
            # for allocation of this replica by using the `is_allocated`
            # method. After that, it calls `reconfigure` to trigger
            # user code initialization.
            async def initialize_replica():
                if is_function:
                    _callable = deployment_def
                else:
                    # This allows deployments to define an async __init__
                    # method (required for FastAPI).
                    _callable = deployment_def.__new__(deployment_def)
                    await sync_to_async(_callable.__init__)(*init_args, **init_kwargs)

                # Setting the context again to update the servable_object.
                ray.serve.context._set_internal_replica_context(
                    deployment_name,
                    replica_tag,
                    controller_name,
                    servable_object=_callable,
                )

                self.replica = RayServeReplica(
                    _callable,
                    deployment_name,
                    replica_tag,
                    deployment_config,
                    deployment_config.user_config,
                    version,
                    is_function,
                    controller_handle,
                )

            # Is it fine that replica is None here?
            # Should we add a check in all methods that use self.replica
            # or, alternatively, create an async get_replica() method?
            self.replica = None
            self._initialize_replica = initialize_replica

        @ray.method(num_returns=2)
        async def handle_request(
            self,
            pickled_request_metadata: bytes,
            *request_args,
            **request_kwargs,
        ):
            # The request metadata should be pickled for performance.
            request_metadata: RequestMetadata = pickle.loads(pickled_request_metadata)

            # Directly receive input because it might contain an ObjectRef.
            query = Query(request_args, request_kwargs, request_metadata)
            return await self.replica.handle_request(query)

        async def handle_request_from_java(
            self,
            proto_request_metadata: bytes,
            *request_args,
            **request_kwargs,
        ):
            from ray.serve.generated.serve_pb2 import (
                RequestMetadata as RequestMetadataProto,
            )

            proto = RequestMetadataProto.FromString(proto_request_metadata)
            request_metadata: RequestMetadata = RequestMetadata(
                proto.request_id, proto.endpoint, call_method=proto.call_method
            )
            request_args = request_args[0]
            query = Query(request_args, request_kwargs, request_metadata, return_num=1)
            return await self.replica.handle_request(query)

        async def is_allocated(self) -> str:
            """poke the replica to check whether it's alive.

            When calling this method on an ActorHandle, it will complete as
            soon as the actor has started running. We use this mechanism to
            detect when a replica has been allocated a worker slot.
            At this time, the replica can transition from PENDING_ALLOCATION
            to PENDING_INITIALIZATION startup state.

            Return the NodeID of this replica
            """
            return ray.get_runtime_context().get_node_id()

        async def is_initialized(
            self, user_config: Optional[Any] = None, _after: Optional[Any] = None
        ):
            # Unused `_after` argument is for scheduling: passing an ObjectRef
            # allows delaying reconfiguration until after this call has returned.
            await self._initialize_replica()

            metadata = await self.reconfigure(user_config)

            # A new replica should not be considered healthy until it passes an
            # initial health check. If an initial health check fails, consider
            # it an initialization failure.
            await self.check_health()
            return metadata

        async def reconfigure(
            self, user_config: Optional[Any] = None
        ) -> Tuple[DeploymentConfig, DeploymentVersion]:
            if user_config is not None:
                await self.replica.reconfigure(user_config)

            return self.get_metadata()

        def get_metadata(self) -> Tuple[DeploymentConfig, DeploymentVersion]:
            return self.replica.deployment_config, self.replica.version

        async def prepare_for_shutdown(self):
            if self.replica is not None:
                return await self.replica.prepare_for_shutdown()

        @ray.method(concurrency_group=HEALTH_CHECK_CONCURRENCY_GROUP)
        async def check_health(self):
            await self.replica.check_health()

    # Dynamically create a new class with custom name here so Ray picks it up
    # correctly in actor metadata table and observability stack.
    return type(
        _format_replica_actor_name(name),
        (RayServeWrappedReplica,),
        dict(RayServeWrappedReplica.__dict__),
    )


class RayServeReplica:
    """Handles requests with the provided callable."""

    def __init__(
        self,
        _callable: Callable,
        deployment_name: str,
        replica_tag: ReplicaTag,
        deployment_config: DeploymentConfig,
        user_config: Any,
        version: DeploymentVersion,
        is_function: bool,
        controller_handle: ActorHandle,
    ) -> None:
        self.deployment_config = deployment_config
        self.deployment_name = deployment_name
        self.replica_tag = replica_tag
        self.callable = _callable
        self.is_function = is_function
        self.user_config = user_config
        self.version = version
        self.rwlock = aiorwlock.RWLock()

        user_health_check = getattr(_callable, HEALTH_CHECK_METHOD, None)
        if not callable(user_health_check):

            def user_health_check():
                pass

        self.user_health_check = sync_to_async(user_health_check)

        self.num_ongoing_requests = 0

        self.request_counter = metrics.Counter(
            "serve_deployment_request_counter",
            description=(
                "The number of queries that have been processed in this replica."
            ),
            tag_keys=("deployment", "replica"),
        )
        self.request_counter.set_default_tags(
            {"deployment": self.deployment_name, "replica": self.replica_tag}
        )

        self.error_counter = metrics.Counter(
            "serve_deployment_error_counter",
            description=(
                "The number of exceptions that have occurred in this replica."
            ),
            tag_keys=("deployment", "replica"),
        )
        self.error_counter.set_default_tags(
            {"deployment": self.deployment_name, "replica": self.replica_tag}
        )

        self.restart_counter = metrics.Counter(
            "serve_deployment_replica_starts",
            description=(
                "The number of times this replica has been restarted due to failure."
            ),
            tag_keys=("deployment", "replica"),
        )
        self.restart_counter.set_default_tags(
            {"deployment": self.deployment_name, "replica": self.replica_tag}
        )

        self.processing_latency_tracker = metrics.Histogram(
            "serve_deployment_processing_latency_ms",
            description="The latency for queries to be processed.",
            boundaries=DEFAULT_LATENCY_BUCKET_MS,
            tag_keys=("deployment", "replica"),
        )
        self.processing_latency_tracker.set_default_tags(
            {"deployment": self.deployment_name, "replica": self.replica_tag}
        )

        self.num_processing_items = metrics.Gauge(
            "serve_replica_processing_queries",
            description="The current number of queries being processed.",
            tag_keys=("deployment", "replica"),
        )
        self.num_processing_items.set_default_tags(
            {"deployment": self.deployment_name, "replica": self.replica_tag}
        )

        self.restart_counter.inc()

        self._shutdown_wait_loop_s = deployment_config.graceful_shutdown_wait_loop_s

        if deployment_config.autoscaling_config:
            process_remote_func = controller_handle.record_autoscaling_metrics.remote
            config = deployment_config.autoscaling_config
            start_metrics_pusher(
                interval_s=config.metrics_interval_s,
                collection_callback=self._collect_autoscaling_metrics,
                metrics_process_func=process_remote_func,
            )

        # NOTE(edoakes): we used to recommend that users use the "ray" logger
        # and tagged the logs with metadata as below. We now recommend using
        # the "ray.serve" 'component logger' (as of Ray 1.13). This is left to
        # maintain backwards compatibility with users who were using the
        # existing logger. We can consider removing it in Ray 2.0.
        ray_logger = logging.getLogger("ray")
        for handler in ray_logger.handlers:
            handler.setFormatter(
                logging.Formatter(
                    handler.formatter._fmt
                    + f" component=serve deployment={self.deployment_name} "
                    f"replica={self.replica_tag}"
                )
            )

    async def check_health(self):
        await self.user_health_check()

    def _get_handle_request_stats(self) -> Optional[Dict[str, int]]:
        actor_stats = ray.runtime_context.get_runtime_context()._get_actor_call_stats()
        method_stat = actor_stats.get(
            f"{_format_replica_actor_name(self.deployment_name)}.handle_request"
        )
        method_stat_java = actor_stats.get(
            f"{_format_replica_actor_name(self.deployment_name)}"
            f".handle_request_from_java"
        )
        return merge_dict(method_stat, method_stat_java)

    def _collect_autoscaling_metrics(self):
        method_stat = self._get_handle_request_stats()

        num_inflight_requests = 0
        if method_stat is not None:
            num_inflight_requests = method_stat["pending"] + method_stat["running"]

        return {self.replica_tag: num_inflight_requests}

    def get_runner_method(self, request_item: Query) -> Callable:
        method_name = request_item.metadata.call_method
        if not hasattr(self.callable, method_name):
            # Filter to methods that don't start with '__' prefix.
            def callable_method_filter(attr):
                if attr.startswith("__"):
                    return False
                elif not callable(getattr(self.callable, attr)):
                    return False

                return True

            methods = list(filter(callable_method_filter, dir(self.callable)))
            raise RayServeException(
                f"Tried to call a method '{method_name}' "
                "that does not exist. Available methods: "
                f"{methods}."
            )
        if self.is_function:
            return self.callable
        return getattr(self.callable, method_name)

    async def ensure_serializable_response(self, response: Any) -> Any:
        if isinstance(response, starlette.responses.StreamingResponse):

            async def mock_receive():
                # This is called in a tight loop in response() just to check
                # for an http disconnect.  So rather than return immediately
                # we should suspend execution to avoid wasting CPU cycles.
                never_set_event = asyncio.Event()
                await never_set_event.wait()

            sender = ASGIHTTPSender()
            await response(scope=None, receive=mock_receive, send=sender)
            return sender.build_asgi_response()
        return response

    async def invoke_single(self, request_item: Query) -> Tuple[Any, bool]:
        """Executes the provided request on this replica.

        Returns the user-provided output and a boolean indicating if the
        request succeeded (user code didn't raise an exception).
        """
        logger.debug(
            "Replica {} started executing request {}".format(
                self.replica_tag, request_item.metadata.request_id
            )
        )
        args, kwargs = parse_request_item(request_item)

        method_to_call = None
        success = True
        try:
            runner_method = self.get_runner_method(request_item)
            method_to_call = sync_to_async(runner_method)
            result = None
            if len(inspect.signature(runner_method).parameters) > 0:
                result = await method_to_call(*args, **kwargs)
            else:
                # When access via http http_arg_is_pickled with no args:
                # args = (<starlette.requests.Request object at 0x7fe900694cc0>,)
                # When access via python with no args:
                # args = ()
                if len(args) == 1 and isinstance(args[0], starlette.requests.Request):
                    # The method doesn't take in anything, including the request
                    # information, so we pass nothing into it
                    result = await method_to_call()
                else:
                    # Will throw due to signature mismatch if user attempts to
                    # call with non-empty args
                    result = await method_to_call(*args, **kwargs)

            result = await self.ensure_serializable_response(result)
            self.request_counter.inc()
        except Exception as e:
            logger.exception(f"Request failed due to {type(e).__name__}:")
            success = False

            # If the debugger is enabled, drop into the remote pdb here.
            if ray.util.pdb._is_ray_debugger_enabled():
                ray.util.pdb._post_mortem()

            function_name = "unknown"
            if method_to_call is not None:
                function_name = method_to_call.__name__
            result = wrap_to_ray_error(function_name, e)
            self.error_counter.inc()

        return result, success

    async def reconfigure(self, user_config: Any):
        async with self.rwlock.writer_lock:
            self.user_config = user_config
            self.version = DeploymentVersion(
                self.version.code_version, user_config=user_config
            )
            if self.is_function:
                raise ValueError("deployment_def must be a class to use user_config")
            elif not hasattr(self.callable, RECONFIGURE_METHOD):
                raise RayServeException(
                    "user_config specified but deployment "
                    + self.deployment_name
                    + " missing "
                    + RECONFIGURE_METHOD
                    + " method"
                )
            reconfigure_method = sync_to_async(
                getattr(self.callable, RECONFIGURE_METHOD)
            )
            await reconfigure_method(user_config)

    async def handle_request(self, request: Query) -> asyncio.Future:
        async with self.rwlock.reader_lock:
            num_running_requests = self._get_handle_request_stats()["running"]
            self.num_processing_items.set(num_running_requests)

            start_time = time.time()
            result, success = await self.invoke_single(request)
            latency_ms = (time.time() - start_time) * 1000

            self.processing_latency_tracker.observe(latency_ms)

            logger.info(
                access_log_msg(
                    method="HANDLE",
                    route=request.metadata.call_method,
                    status="OK" if success else "ERROR",
                    latency_ms=latency_ms,
                )
            )
            if request.return_num == 1:
                return result
            else:
                # Returns a small object for router to track request status.
                return b"", result

    async def prepare_for_shutdown(self):
        """Perform graceful shutdown.

        Trigger a graceful shutdown protocol that will wait for all the queued
        tasks to be completed and return to the controller.
        """
        while True:
            # Sleep first because we want to make sure all the routers receive
            # the notification to remove this replica first.
            await asyncio.sleep(self._shutdown_wait_loop_s)
            method_stat = self._get_handle_request_stats()
            # The handle_request method wasn't even invoked.
            if method_stat is None:
                break
            # The handle_request method has 0 inflight requests.
            if method_stat["running"] + method_stat["pending"] == 0:
                break
            else:
                logger.info(
                    "Waiting for an additional "
                    f"{self._shutdown_wait_loop_s}s to shut down because "
                    f"there are {self.num_ongoing_requests} ongoing requests."
                )

        # Explicitly call the del method to trigger clean up.
        # We set the del method to noop after successfully calling it so the
        # destructor is called only once.
        try:
            if hasattr(self.callable, "__del__"):
                # Make sure to accept `async def __del__(self)` as well.
                await sync_to_async(self.callable.__del__)()
                setattr(self.callable, "__del__", lambda _: None)
        except Exception as e:
            logger.exception(f"Exception during graceful shutdown of replica: {e}")
        finally:
            if hasattr(self.callable, "__del__"):
                del self.callable



import os
from typing import TYPE_CHECKING, Optional

from sklearn.base import BaseEstimator
from ray.air._internal.checkpointing import save_preprocessor_to_dir
from ray.air.checkpoint import Checkpoint
from ray.air.constants import MODEL_KEY
import ray.cloudpickle as cpickle
from ray.util.annotations import PublicAPI

if TYPE_CHECKING:
    from ray.data.preprocessor import Preprocessor


@PublicAPI(stability="alpha")
class SklearnCheckpoint(Checkpoint):
    """A :py:class:`~ray.air.checkpoint.Checkpoint` with sklearn-specific
    functionality.

    Create this from a generic :py:class:`~ray.air.checkpoint.Checkpoint` by calling
    ``SklearnCheckpoint.from_checkpoint(ckpt)``
    """

    @classmethod
    def from_estimator(
        cls,
        estimator: BaseEstimator,
        *,
        path: os.PathLike,
        preprocessor: Optional["Preprocessor"] = None,
    ) -> "SklearnCheckpoint":
        """Create a :py:class:`~ray.air.checkpoint.Checkpoint` that stores an sklearn
        ``Estimator``.

        Args:
            estimator: The ``Estimator`` to store in the checkpoint.
            path: The directory where the checkpoint will be stored.
            preprocessor: A fitted preprocessor to be applied before inference.

        Returns:
            An :py:class:`SklearnCheckpoint` containing the specified ``Estimator``.

        Examples:
            >>> from ray.train.sklearn import SklearnCheckpoint
            >>> from sklearn.ensemble import RandomForestClassifier
            >>>
            >>> estimator = RandomForestClassifier()
            >>> checkpoint = SklearnCheckpoint.from_estimator(estimator, path=".")

            You can use a :py:class:`SklearnCheckpoint` to create an
            :py:class:`~ray.train.sklearn.SklearnPredictor` and preform inference.

            >>> from ray.train.sklearn import SklearnPredictor
            >>>
            >>> predictor = SklearnPredictor.from_checkpoint(checkpoint)
        """
        with open(os.path.join(path, MODEL_KEY), "wb") as f:
            cpickle.dump(estimator, f)

        if preprocessor:
            save_preprocessor_to_dir(preprocessor, path)

        checkpoint = cls.from_directory(path)

        return checkpoint

    def get_estimator(self) -> BaseEstimator:
        """Retrieve the ``Estimator`` stored in this checkpoint."""
        with self.as_directory() as checkpoint_path:
            estimator_path = os.path.join(checkpoint_path, MODEL_KEY)
            with open(estimator_path, "rb") as f:
                return cpickle.load(f)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: src/ray/protobuf/gcs.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import enum_type_wrapper
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from . import common_pb2 as src_dot_ray_dot_protobuf_dot_common__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x1asrc/ray/protobuf/gcs.proto\x12\x07ray.rpc\x1a\x1dsrc/ray/protobuf/common.proto\"m\n\x08GcsEntry\x12\x37\n\x0b\x63hange_mode\x18\x01 \x01(\x0e\x32\x16.ray.rpc.GcsChangeModeR\nchangeMode\x12\x0e\n\x02id\x18\x02 \x01(\x0cR\x02id\x12\x18\n\x07\x65ntries\x18\x03 \x03(\x0cR\x07\x65ntries\"+\n\x0fObjectTableData\x12\x18\n\x07manager\x18\x01 \x01(\x0cR\x07manager\"\xea\t\n\x0e\x41\x63torTableData\x12\x19\n\x08\x61\x63tor_id\x18\x01 \x01(\x0cR\x07\x61\x63torId\x12\x1b\n\tparent_id\x18\x02 \x01(\x0cR\x08parentId\x12\x42\n\x1e\x61\x63tor_creation_dummy_object_id\x18\x03 \x01(\x0cR\x1a\x61\x63torCreationDummyObjectId\x12\x15\n\x06job_id\x18\x04 \x01(\x0cR\x05jobId\x12\x38\n\x05state\x18\x06 \x01(\x0e\x32\".ray.rpc.ActorTableData.ActorStateR\x05state\x12!\n\x0cmax_restarts\x18\x07 \x01(\x03R\x0bmaxRestarts\x12!\n\x0cnum_restarts\x18\x08 \x01(\x04R\x0bnumRestarts\x12*\n\x07\x61\x64\x64ress\x18\t \x01(\x0b\x32\x10.ray.rpc.AddressR\x07\x61\x64\x64ress\x12\x35\n\rowner_address\x18\n \x01(\x0b\x32\x10.ray.rpc.AddressR\x0cownerAddress\x12\x1f\n\x0bis_detached\x18\x0b \x01(\x08R\nisDetached\x12\x12\n\x04name\x18\x0c \x01(\tR\x04name\x12\x1c\n\ttimestamp\x18\r \x01(\x01R\ttimestamp\x12\x44\n\x10resource_mapping\x18\x0f \x03(\x0b\x32\x19.ray.rpc.ResourceMapEntryR\x0fresourceMapping\x12\x10\n\x03pid\x18\x10 \x01(\rR\x03pid\x12L\n\x13\x66unction_descriptor\x18\x11 \x01(\x0b\x32\x1b.ray.rpc.FunctionDescriptorR\x12\x66unctionDescriptor\x12#\n\rray_namespace\x18\x13 \x01(\tR\x0crayNamespace\x12\x1d\n\nstart_time\x18\x14 \x01(\x04R\tstartTime\x12\x19\n\x08\x65nd_time\x18\x15 \x01(\x04R\x07\x65ndTime\x12\x34\n\x16serialized_runtime_env\x18\x16 \x01(\tR\x14serializedRuntimeEnv\x12\x1d\n\nclass_name\x18\x17 \x01(\tR\tclassName\x12\x39\n\x0b\x64\x65\x61th_cause\x18\x18 \x01(\x0b\x32\x18.ray.rpc.ActorDeathCauseR\ndeathCause\x12]\n\x12required_resources\x18\x1c \x03(\x0b\x32..ray.rpc.ActorTableData.RequiredResourcesEntryR\x11requiredResources\x12\x1c\n\x07node_id\x18\x1d \x01(\x0cH\x00R\x06nodeId\x88\x01\x01\x12\x31\n\x12placement_group_id\x18\x1e \x01(\x0cH\x01R\x10placementGroupId\x88\x01\x01\x1a\x44\n\x16RequiredResourcesEntry\x12\x10\n\x03key\x18\x01 \x01(\tR\x03key\x12\x14\n\x05value\x18\x02 \x01(\x01R\x05value:\x02\x38\x01\"a\n\nActorState\x12\x18\n\x14\x44\x45PENDENCIES_UNREADY\x10\x00\x12\x14\n\x10PENDING_CREATION\x10\x01\x12\t\n\x05\x41LIVE\x10\x02\x12\x0e\n\nRESTARTING\x10\x03\x12\x08\n\x04\x44\x45\x41\x44\x10\x04\x42\n\n\x08_node_idB\x15\n\x13_placement_group_id\"~\n\x0e\x45rrorTableData\x12\x15\n\x06job_id\x18\x01 \x01(\x0cR\x05jobId\x12\x12\n\x04type\x18\x02 \x01(\tR\x04type\x12#\n\rerror_message\x18\x03 \x01(\tR\x0c\x65rrorMessage\x12\x1c\n\ttimestamp\x18\x04 \x01(\x01R\ttimestamp\"\x9d\x01\n\x0cScheduleData\x12L\n\rschedule_plan\x18\x01 \x03(\x0b\x32\'.ray.rpc.ScheduleData.SchedulePlanEntryR\x0cschedulePlan\x1a?\n\x11SchedulePlanEntry\x12\x10\n\x03key\x18\x01 \x01(\tR\x03key\x12\x14\n\x05value\x18\x02 \x01(\x0cR\x05value:\x02\x38\x01\"\x9f\x01\n\x11ProfileEventEntry\x12\x1d\n\nstart_time\x18\x01 \x01(\x03R\tstartTime\x12\x19\n\x08\x65nd_time\x18\x02 \x01(\x03R\x07\x65ndTime\x12\"\n\nextra_data\x18\x03 \x01(\tH\x00R\textraData\x88\x01\x01\x12\x1d\n\nevent_name\x18\x04 \x01(\tR\teventNameB\r\n\x0b_extra_data\"\xb5\x01\n\rProfileEvents\x12%\n\x0e\x63omponent_type\x18\x01 \x01(\tR\rcomponentType\x12!\n\x0c\x63omponent_id\x18\x02 \x01(\x0cR\x0b\x63omponentId\x12&\n\x0fnode_ip_address\x18\x03 \x01(\tR\rnodeIpAddress\x12\x32\n\x06\x65vents\x18\x04 \x03(\x0b\x32\x1a.ray.rpc.ProfileEventEntryR\x06\x65vents\"\x8c\x04\n\x0fTaskStateUpdate\x12\x1c\n\x07node_id\x18\x01 \x01(\x0cH\x00R\x06nodeId\x88\x01\x01\x12\x36\n\x15pending_args_avail_ts\x18\x02 \x01(\x03H\x01R\x12pendingArgsAvailTs\x88\x01\x01\x12@\n\x1apending_node_assignment_ts\x18\x03 \x01(\x03H\x02R\x17pendingNodeAssignmentTs\x88\x01\x01\x12\x38\n\x16submitted_to_worker_ts\x18\x04 \x01(\x03H\x03R\x13submittedToWorkerTs\x88\x01\x01\x12\"\n\nrunning_ts\x18\x05 \x01(\x03H\x04R\trunningTs\x88\x01\x01\x12$\n\x0b\x66inished_ts\x18\x06 \x01(\x03H\x05R\nfinishedTs\x88\x01\x01\x12 \n\tfailed_ts\x18\x07 \x01(\x03H\x06R\x08\x66\x61iledTs\x88\x01\x01\x12 \n\tworker_id\x18\x08 \x01(\x0cH\x07R\x08workerId\x88\x01\x01\x42\n\n\x08_node_idB\x18\n\x16_pending_args_avail_tsB\x1d\n\x1b_pending_node_assignment_tsB\x19\n\x17_submitted_to_worker_tsB\r\n\x0b_running_tsB\x0e\n\x0c_finished_tsB\x0c\n\n_failed_tsB\x0c\n\n_worker_id\"\xd8\x02\n\nTaskEvents\x12\x17\n\x07task_id\x18\x01 \x01(\x0cR\x06taskId\x12%\n\x0e\x61ttempt_number\x18\x02 \x01(\x05R\rattemptNumber\x12\x38\n\ttask_info\x18\x03 \x01(\x0b\x32\x16.ray.rpc.TaskInfoEntryH\x00R\x08taskInfo\x88\x01\x01\x12\x42\n\rstate_updates\x18\x04 \x01(\x0b\x32\x18.ray.rpc.TaskStateUpdateH\x01R\x0cstateUpdates\x88\x01\x01\x12\x42\n\x0eprofile_events\x18\x05 \x01(\x0b\x32\x16.ray.rpc.ProfileEventsH\x02R\rprofileEvents\x88\x01\x01\x12\x15\n\x06job_id\x18\x06 \x01(\x0cR\x05jobIdB\x0c\n\n_task_infoB\x10\n\x0e_state_updatesB\x11\n\x0f_profile_events\"\xd4\x01\n\rTaskEventData\x12\x39\n\x0e\x65vents_by_task\x18\x01 \x03(\x0b\x32\x13.ray.rpc.TaskEventsR\x0c\x65ventsByTask\x12\x44\n\x1fnum_profile_task_events_dropped\x18\x03 \x01(\x05R\x1bnumProfileTaskEventsDropped\x12\x42\n\x1enum_status_task_events_dropped\x18\x04 \x01(\x05R\x1anumStatusTaskEventsDropped\"@\n\x11ResourceTableData\x12+\n\x11resource_capacity\x18\x01 \x01(\x01R\x10resourceCapacity\"\xda\x01\n\x12\x41vailableResources\x12\x17\n\x07node_id\x18\x01 \x01(\x0cR\x06nodeId\x12\x64\n\x13resources_available\x18\x02 \x03(\x0b\x32\x33.ray.rpc.AvailableResources.ResourcesAvailableEntryR\x12resourcesAvailable\x1a\x45\n\x17ResourcesAvailableEntry\x12\x10\n\x03key\x18\x01 \x01(\tR\x03key\x12\x14\n\x05value\x18\x02 \x01(\x01R\x05value:\x02\x38\x01\"\xd4\x05\n\x0bGcsNodeInfo\x12\x17\n\x07node_id\x18\x01 \x01(\x0cR\x06nodeId\x12\x30\n\x14node_manager_address\x18\x02 \x01(\tR\x12nodeManagerAddress\x12,\n\x12raylet_socket_name\x18\x03 \x01(\tR\x10rayletSocketName\x12\x37\n\x18object_store_socket_name\x18\x04 \x01(\tR\x15objectStoreSocketName\x12*\n\x11node_manager_port\x18\x05 \x01(\x05R\x0fnodeManagerPort\x12.\n\x13object_manager_port\x18\x06 \x01(\x05R\x11objectManagerPort\x12\x37\n\x05state\x18\x07 \x01(\x0e\x32!.ray.rpc.GcsNodeInfo.GcsNodeStateR\x05state\x12\x32\n\x15node_manager_hostname\x18\x08 \x01(\tR\x13nodeManagerHostname\x12.\n\x13metrics_export_port\x18\t \x01(\x05R\x11metricsExportPort\x12Q\n\x0fresources_total\x18\x0b \x03(\x0b\x32(.ray.rpc.GcsNodeInfo.ResourcesTotalEntryR\x0eresourcesTotal\x12\x1b\n\tnode_name\x18\x0c \x01(\tR\x08nodeName\x12\"\n\rstart_time_ms\x18\x17 \x01(\x04R\x0bstartTimeMs\x12\x1e\n\x0b\x65nd_time_ms\x18\x18 \x01(\x04R\tendTimeMs\x1a\x41\n\x13ResourcesTotalEntry\x12\x10\n\x03key\x18\x01 \x01(\tR\x03key\x12\x14\n\x05value\x18\x02 \x01(\x01R\x05value:\x02\x38\x01\"#\n\x0cGcsNodeState\x12\t\n\x05\x41LIVE\x10\x00\x12\x08\n\x04\x44\x45\x41\x44\x10\x01\"\xa4\x07\n\x0bJobsAPIInfo\x12\x16\n\x06status\x18\x01 \x01(\tR\x06status\x12\x1e\n\nentrypoint\x18\x02 \x01(\tR\nentrypoint\x12\x1d\n\x07message\x18\x03 \x01(\tH\x00R\x07message\x88\x01\x01\x12\"\n\nerror_type\x18\x04 \x01(\tH\x01R\terrorType\x88\x01\x01\x12\"\n\nstart_time\x18\x05 \x01(\x04H\x02R\tstartTime\x88\x01\x01\x12\x1e\n\x08\x65nd_time\x18\x06 \x01(\x04H\x03R\x07\x65ndTime\x88\x01\x01\x12>\n\x08metadata\x18\x07 \x03(\x0b\x32\".ray.rpc.JobsAPIInfo.MetadataEntryR\x08metadata\x12-\n\x10runtime_env_json\x18\x08 \x01(\tH\x04R\x0eruntimeEnvJson\x88\x01\x01\x12\x33\n\x13\x65ntrypoint_num_cpus\x18\t \x01(\x01H\x05R\x11\x65ntrypointNumCpus\x88\x01\x01\x12\x33\n\x13\x65ntrypoint_num_gpus\x18\n \x01(\x01H\x06R\x11\x65ntrypointNumGpus\x88\x01\x01\x12`\n\x14\x65ntrypoint_resources\x18\x0b \x03(\x0b\x32-.ray.rpc.JobsAPIInfo.EntrypointResourcesEntryR\x13\x65ntrypointResources\x12>\n\x19\x64river_agent_http_address\x18\x0c \x01(\tH\x07R\x16\x64riverAgentHttpAddress\x88\x01\x01\x12)\n\x0e\x64river_node_id\x18\r \x01(\tH\x08R\x0c\x64riverNodeId\x88\x01\x01\x1a;\n\rMetadataEntry\x12\x10\n\x03key\x18\x01 \x01(\tR\x03key\x12\x14\n\x05value\x18\x02 \x01(\tR\x05value:\x02\x38\x01\x1a\x46\n\x18\x45ntrypointResourcesEntry\x12\x10\n\x03key\x18\x01 \x01(\tR\x03key\x12\x14\n\x05value\x18\x02 \x01(\x01R\x05value:\x02\x38\x01\x42\n\n\x08_messageB\r\n\x0b_error_typeB\r\n\x0b_start_timeB\x0b\n\t_end_timeB\x13\n\x11_runtime_env_jsonB\x16\n\x14_entrypoint_num_cpusB\x16\n\x14_entrypoint_num_gpusB\x1c\n\x1a_driver_agent_http_addressB\x11\n\x0f_driver_node_id\"\xf0\x02\n\x0cJobTableData\x12\x15\n\x06job_id\x18\x01 \x01(\x0cR\x05jobId\x12\x17\n\x07is_dead\x18\x02 \x01(\x08R\x06isDead\x12\x1c\n\ttimestamp\x18\x03 \x01(\x03R\ttimestamp\x12*\n\x11\x64river_ip_address\x18\x04 \x01(\tR\x0f\x64riverIpAddress\x12\x1d\n\ndriver_pid\x18\x05 \x01(\x03R\tdriverPid\x12*\n\x06\x63onfig\x18\x06 \x01(\x0b\x32\x12.ray.rpc.JobConfigR\x06\x63onfig\x12\x1d\n\nstart_time\x18\x07 \x01(\x04R\tstartTime\x12\x19\n\x08\x65nd_time\x18\x08 \x01(\x04R\x07\x65ndTime\x12\x1e\n\nentrypoint\x18\t \x01(\tR\nentrypoint\x12\x34\n\x08job_info\x18\n \x01(\x0b\x32\x14.ray.rpc.JobsAPIInfoH\x00R\x07jobInfo\x88\x01\x01\x42\x0b\n\t_job_info\"\xe7\x04\n\x0fWorkerTableData\x12\x19\n\x08is_alive\x18\x01 \x01(\x08R\x07isAlive\x12\x37\n\x0eworker_address\x18\x02 \x01(\x0b\x32\x10.ray.rpc.AddressR\rworkerAddress\x12\x1c\n\ttimestamp\x18\x03 \x01(\x03R\ttimestamp\x12\x34\n\x0bworker_type\x18\x05 \x01(\x0e\x32\x13.ray.rpc.WorkerTypeR\nworkerType\x12I\n\x0bworker_info\x18\x06 \x03(\x0b\x32(.ray.rpc.WorkerTableData.WorkerInfoEntryR\nworkerInfo\x12M\n\x17\x63reation_task_exception\x18\x12 \x01(\x0b\x32\x15.ray.rpc.RayExceptionR\x15\x63reationTaskException\x12\x39\n\texit_type\x18\x13 \x01(\x0e\x32\x17.ray.rpc.WorkerExitTypeH\x00R\x08\x65xitType\x88\x01\x01\x12$\n\x0b\x65xit_detail\x18\x14 \x01(\tH\x01R\nexitDetail\x88\x01\x01\x12\x10\n\x03pid\x18\x15 \x01(\rR\x03pid\x12\"\n\rstart_time_ms\x18\x17 \x01(\x04R\x0bstartTimeMs\x12\x1e\n\x0b\x65nd_time_ms\x18\x18 \x01(\x04R\tendTimeMs\x1a=\n\x0fWorkerInfoEntry\x12\x10\n\x03key\x18\x01 \x01(\tR\x03key\x12\x14\n\x05value\x18\x02 \x01(\x0cR\x05value:\x02\x38\x01\x42\x0c\n\n_exit_typeB\x0e\n\x0c_exit_detail\"K\n\x0fWorkerDeltaData\x12\x1b\n\traylet_id\x18\x01 \x01(\x0cR\x08rayletId\x12\x1b\n\tworker_id\x18\x02 \x01(\x0cR\x08workerId\"\x9a\x01\n\x0bResourceMap\x12\x35\n\x05items\x18\x01 \x03(\x0b\x32\x1f.ray.rpc.ResourceMap.ItemsEntryR\x05items\x1aT\n\nItemsEntry\x12\x10\n\x03key\x18\x01 \x01(\tR\x03key\x12\x30\n\x05value\x18\x02 \x01(\x0b\x32\x1a.ray.rpc.ResourceTableDataR\x05value:\x02\x38\x01\"&\n\x0cStoredConfig\x12\x16\n\x06\x63onfig\x18\x01 \x01(\tR\x06\x63onfig\"\xff\x01\n\x12NodeResourceChange\x12\x17\n\x07node_id\x18\x01 \x01(\x0cR\x06nodeId\x12^\n\x11updated_resources\x18\x02 \x03(\x0b\x32\x31.ray.rpc.NodeResourceChange.UpdatedResourcesEntryR\x10updatedResources\x12+\n\x11\x64\x65leted_resources\x18\x03 \x03(\tR\x10\x64\x65letedResources\x1a\x43\n\x15UpdatedResourcesEntry\x12\x10\n\x03key\x18\x01 \x01(\tR\x03key\x12\x14\n\x05value\x18\x02 \x01(\x01R\x05value:\x02\x38\x01\"3\n\rPubSubMessage\x12\x0e\n\x02id\x18\x01 \x01(\x0cR\x02id\x12\x12\n\x04\x64\x61ta\x18\x02 \x01(\x0cR\x04\x64\x61ta\"\x90\x01\n\x0eResourceUpdate\x12\x35\n\x06\x63hange\x18\x01 \x01(\x0b\x32\x1b.ray.rpc.NodeResourceChangeH\x00R\x06\x63hange\x12,\n\x04\x64\x61ta\x18\x02 \x01(\x0b\x32\x16.ray.rpc.ResourcesDataH\x00R\x04\x64\x61taB\x19\n\x17resource_change_or_data\"b\n\x1aResourceUsageBroadcastData\x12\x15\n\x06seq_no\x18\x01 \x01(\x03R\x05seqNo\x12-\n\x05\x62\x61tch\x18\x02 \x03(\x0b\x32\x17.ray.rpc.ResourceUpdateR\x05\x62\x61tch\"\xa7\x02\n\x0eResourceDemand\x12\x38\n\x05shape\x18\x01 \x03(\x0b\x32\".ray.rpc.ResourceDemand.ShapeEntryR\x05shape\x12\x39\n\x19num_ready_requests_queued\x18\x02 \x01(\x04R\x16numReadyRequestsQueued\x12\x43\n\x1enum_infeasible_requests_queued\x18\x03 \x01(\x04R\x1bnumInfeasibleRequestsQueued\x12!\n\x0c\x62\x61\x63klog_size\x18\x04 \x01(\x03R\x0b\x62\x61\x63klogSize\x1a\x38\n\nShapeEntry\x12\x10\n\x03key\x18\x01 \x01(\tR\x03key\x12\x14\n\x05value\x18\x02 \x01(\x01R\x05value:\x02\x38\x01\"R\n\x0cResourceLoad\x12\x42\n\x10resource_demands\x18\x01 \x03(\x0b\x32\x17.ray.rpc.ResourceDemandR\x0fresourceDemands\"\xc1\t\n\rResourcesData\x12\x17\n\x07node_id\x18\x01 \x01(\x0cR\x06nodeId\x12_\n\x13resources_available\x18\x02 \x03(\x0b\x32..ray.rpc.ResourcesData.ResourcesAvailableEntryR\x12resourcesAvailable\x12>\n\x1bresources_available_changed\x18\x03 \x01(\x08R\x19resourcesAvailableChanged\x12S\n\x0fresources_total\x18\x04 \x03(\x0b\x32*.ray.rpc.ResourcesData.ResourcesTotalEntryR\x0eresourcesTotal\x12M\n\rresource_load\x18\x05 \x03(\x0b\x32(.ray.rpc.ResourcesData.ResourceLoadEntryR\x0cresourceLoad\x12\x32\n\x15resource_load_changed\x18\x06 \x01(\x08R\x13resourceLoadChanged\x12J\n\x16resource_load_by_shape\x18\x07 \x01(\x0b\x32\x15.ray.rpc.ResourceLoadR\x13resourceLoadByShape\x12(\n\x10should_global_gc\x18\x08 \x01(\x08R\x0eshouldGlobalGc\x12\x30\n\x14node_manager_address\x18\t \x01(\tR\x12nodeManagerAddress\x12.\n\x13object_pulls_queued\x18\n \x01(\x08R\x11objectPullsQueued\x12\x63\n\x15resources_normal_task\x18\x0b \x03(\x0b\x32/.ray.rpc.ResourcesData.ResourcesNormalTaskEntryR\x13resourcesNormalTask\x12\x41\n\x1dresources_normal_task_changed\x18\x0c \x01(\x08R\x1aresourcesNormalTaskChanged\x12\x45\n\x1fresources_normal_task_timestamp\x18\r \x01(\x03R\x1cresourcesNormalTaskTimestamp\x12\x44\n\x1f\x63luster_full_of_actors_detected\x18\x0e \x01(\x08R\x1b\x63lusterFullOfActorsDetected\x1a\x45\n\x17ResourcesAvailableEntry\x12\x10\n\x03key\x18\x01 \x01(\tR\x03key\x12\x14\n\x05value\x18\x02 \x01(\x01R\x05value:\x02\x38\x01\x1a\x41\n\x13ResourcesTotalEntry\x12\x10\n\x03key\x18\x01 \x01(\tR\x03key\x12\x14\n\x05value\x18\x02 \x01(\x01R\x05value:\x02\x38\x01\x1a?\n\x11ResourceLoadEntry\x12\x10\n\x03key\x18\x01 \x01(\tR\x03key\x12\x14\n\x05value\x18\x02 \x01(\x01R\x05value:\x02\x38\x01\x1a\x46\n\x18ResourcesNormalTaskEntry\x12\x10\n\x03key\x18\x01 \x01(\tR\x03key\x12\x14\n\x05value\x18\x02 \x01(\x01R\x05value:\x02\x38\x01\"\xe1\x01\n\x16ResourceUsageBatchData\x12,\n\x05\x62\x61tch\x18\x01 \x03(\x0b\x32\x16.ray.rpc.ResourcesDataR\x05\x62\x61tch\x12J\n\x16resource_load_by_shape\x18\x02 \x01(\x0b\x32\x15.ray.rpc.ResourceLoadR\x13resourceLoadByShape\x12M\n\x14placement_group_load\x18\x03 \x01(\x0b\x32\x1b.ray.rpc.PlacementGroupLoadR\x12placementGroupLoad\"h\n\x12PlacementGroupLoad\x12R\n\x14placement_group_data\x18\x01 \x03(\x0b\x32 .ray.rpc.PlacementGroupTableDataR\x12placementGroupData\"\xdc\x04\n\x13PlacementGroupStats\x12?\n\x1c\x63reation_request_received_ns\x18\x02 \x01(\x03R\x19\x63reationRequestReceivedNs\x12;\n\x1ascheduling_started_time_ns\x18\x03 \x01(\x03R\x17schedulingStartedTimeNs\x12\x32\n\x15scheduling_latency_us\x18\x04 \x01(\x03R\x13schedulingLatencyUs\x12\x41\n\x1e\x65nd_to_end_creation_latency_us\x18\x05 \x01(\x03R\x19\x65ndToEndCreationLatencyUs\x12-\n\x12scheduling_attempt\x18\x06 \x01(\rR\x11schedulingAttempt\x12\x33\n\x16highest_retry_delay_ms\x18\x07 \x01(\x01R\x13highestRetryDelayMs\x12W\n\x10scheduling_state\x18\x08 \x01(\x0e\x32,.ray.rpc.PlacementGroupStats.SchedulingStateR\x0fschedulingState\"\x92\x01\n\x0fSchedulingState\x12\n\n\x06QUEUED\x10\x00\x12\x0b\n\x07REMOVED\x10\x01\x12\x16\n\x12SCHEDULING_STARTED\x10\x02\x12\x10\n\x0cNO_RESOURCES\x10\x03\x12\x0e\n\nINFEASIBLE\x10\x04\x12\x1e\n\x1a\x46\x41ILED_TO_COMMIT_RESOURCES\x10\x05\x12\x0c\n\x08\x46INISHED\x10\x06\"\xb6\x05\n\x17PlacementGroupTableData\x12,\n\x12placement_group_id\x18\x01 \x01(\x0cR\x10placementGroupId\x12\x12\n\x04name\x18\x02 \x01(\tR\x04name\x12)\n\x07\x62undles\x18\x03 \x03(\x0b\x32\x0f.ray.rpc.BundleR\x07\x62undles\x12\x36\n\x08strategy\x18\x04 \x01(\x0e\x32\x1a.ray.rpc.PlacementStrategyR\x08strategy\x12J\n\x05state\x18\x05 \x01(\x0e\x32\x34.ray.rpc.PlacementGroupTableData.PlacementGroupStateR\x05state\x12$\n\x0e\x63reator_job_id\x18\x06 \x01(\x0cR\x0c\x63reatorJobId\x12(\n\x10\x63reator_actor_id\x18\x07 \x01(\x0cR\x0e\x63reatorActorId\x12(\n\x10\x63reator_job_dead\x18\x08 \x01(\x08R\x0e\x63reatorJobDead\x12,\n\x12\x63reator_actor_dead\x18\t \x01(\x08R\x10\x63reatorActorDead\x12\x1f\n\x0bis_detached\x18\n \x01(\x08R\nisDetached\x12#\n\rray_namespace\x18\x0b \x01(\tR\x0crayNamespace\x12\x32\n\x05stats\x18\x0c \x01(\x0b\x32\x1c.ray.rpc.PlacementGroupStatsR\x05stats\x12\x38\n\x19max_cpu_fraction_per_node\x18\r \x01(\x01R\x15maxCpuFractionPerNode\"N\n\x13PlacementGroupState\x12\x0b\n\x07PENDING\x10\x00\x12\x0b\n\x07\x43REATED\x10\x01\x12\x0b\n\x07REMOVED\x10\x02\x12\x10\n\x0cRESCHEDULING\x10\x03*\xdc\x02\n\x0bTablePrefix\x12\x14\n\x10TABLE_PREFIX_MIN\x10\x00\x12\n\n\x06UNUSED\x10\x01\x12\x08\n\x04TASK\x10\x02\x12\x0f\n\x0bRAYLET_TASK\x10\x03\x12\x08\n\x04NODE\x10\x04\x12\n\n\x06OBJECT\x10\x05\x12\t\n\x05\x41\x43TOR\x10\x06\x12\x0c\n\x08\x46UNCTION\x10\x07\x12\x17\n\x13TASK_RECONSTRUCTION\x10\x08\x12\x18\n\x14RESOURCE_USAGE_BATCH\x10\t\x12\x07\n\x03JOB\x10\n\x12\x0e\n\nTASK_LEASE\x10\x0c\x12\x11\n\rNODE_RESOURCE\x10\r\x12\x10\n\x0c\x44IRECT_ACTOR\x10\x0e\x12\x0b\n\x07WORKERS\x10\x0f\x12\x13\n\x0fINTERNAL_CONFIG\x10\x10\x12\x1c\n\x18PLACEMENT_GROUP_SCHEDULE\x10\x11\x12\x13\n\x0fPLACEMENT_GROUP\x10\x12\x12\x06\n\x02KV\x10\x13\x12\x13\n\x0f\x41\x43TOR_TASK_SPEC\x10\x14*\xbe\x02\n\x0bTablePubsub\x12\x14\n\x10TABLE_PUBSUB_MIN\x10\x00\x12\x0e\n\nNO_PUBLISH\x10\x01\x12\x0f\n\x0bTASK_PUBSUB\x10\x02\x12\x16\n\x12RAYLET_TASK_PUBSUB\x10\x03\x12\x0f\n\x0bNODE_PUBSUB\x10\x04\x12\x11\n\rOBJECT_PUBSUB\x10\x05\x12\x10\n\x0c\x41\x43TOR_PUBSUB\x10\x06\x12\x1f\n\x1bRESOURCE_USAGE_BATCH_PUBSUB\x10\x07\x12\x15\n\x11TASK_LEASE_PUBSUB\x10\x08\x12\x0e\n\nJOB_PUBSUB\x10\t\x12\x18\n\x14NODE_RESOURCE_PUBSUB\x10\n\x12\x17\n\x13\x44IRECT_ACTOR_PUBSUB\x10\x0b\x12\x19\n\x15WORKER_FAILURE_PUBSUB\x10\x0c\x12\x14\n\x10TABLE_PUBSUB_MAX\x10\r*.\n\rGcsChangeMode\x12\x11\n\rAPPEND_OR_ADD\x10\x00\x12\n\n\x06REMOVE\x10\x01\x42\x1d\n\x18io.ray.runtime.generated\xf8\x01\x01\x62\x06proto3')

_TABLEPREFIX = DESCRIPTOR.enum_types_by_name['TablePrefix']
TablePrefix = enum_type_wrapper.EnumTypeWrapper(_TABLEPREFIX)
_TABLEPUBSUB = DESCRIPTOR.enum_types_by_name['TablePubsub']
TablePubsub = enum_type_wrapper.EnumTypeWrapper(_TABLEPUBSUB)
_GCSCHANGEMODE = DESCRIPTOR.enum_types_by_name['GcsChangeMode']
GcsChangeMode = enum_type_wrapper.EnumTypeWrapper(_GCSCHANGEMODE)
TABLE_PREFIX_MIN = 0
UNUSED = 1
TASK = 2
RAYLET_TASK = 3
NODE = 4
OBJECT = 5
ACTOR = 6
FUNCTION = 7
TASK_RECONSTRUCTION = 8
RESOURCE_USAGE_BATCH = 9
JOB = 10
TASK_LEASE = 12
NODE_RESOURCE = 13
DIRECT_ACTOR = 14
WORKERS = 15
INTERNAL_CONFIG = 16
PLACEMENT_GROUP_SCHEDULE = 17
PLACEMENT_GROUP = 18
KV = 19
ACTOR_TASK_SPEC = 20
TABLE_PUBSUB_MIN = 0
NO_PUBLISH = 1
TASK_PUBSUB = 2
RAYLET_TASK_PUBSUB = 3
NODE_PUBSUB = 4
OBJECT_PUBSUB = 5
ACTOR_PUBSUB = 6
RESOURCE_USAGE_BATCH_PUBSUB = 7
TASK_LEASE_PUBSUB = 8
JOB_PUBSUB = 9
NODE_RESOURCE_PUBSUB = 10
DIRECT_ACTOR_PUBSUB = 11
WORKER_FAILURE_PUBSUB = 12
TABLE_PUBSUB_MAX = 13
APPEND_OR_ADD = 0
REMOVE = 1


_GCSENTRY = DESCRIPTOR.message_types_by_name['GcsEntry']
_OBJECTTABLEDATA = DESCRIPTOR.message_types_by_name['ObjectTableData']
_ACTORTABLEDATA = DESCRIPTOR.message_types_by_name['ActorTableData']
_ACTORTABLEDATA_REQUIREDRESOURCESENTRY = _ACTORTABLEDATA.nested_types_by_name['RequiredResourcesEntry']
_ERRORTABLEDATA = DESCRIPTOR.message_types_by_name['ErrorTableData']
_SCHEDULEDATA = DESCRIPTOR.message_types_by_name['ScheduleData']
_SCHEDULEDATA_SCHEDULEPLANENTRY = _SCHEDULEDATA.nested_types_by_name['SchedulePlanEntry']
_PROFILEEVENTENTRY = DESCRIPTOR.message_types_by_name['ProfileEventEntry']
_PROFILEEVENTS = DESCRIPTOR.message_types_by_name['ProfileEvents']
_TASKSTATEUPDATE = DESCRIPTOR.message_types_by_name['TaskStateUpdate']
_TASKEVENTS = DESCRIPTOR.message_types_by_name['TaskEvents']
_TASKEVENTDATA = DESCRIPTOR.message_types_by_name['TaskEventData']
_RESOURCETABLEDATA = DESCRIPTOR.message_types_by_name['ResourceTableData']
_AVAILABLERESOURCES = DESCRIPTOR.message_types_by_name['AvailableResources']
_AVAILABLERESOURCES_RESOURCESAVAILABLEENTRY = _AVAILABLERESOURCES.nested_types_by_name['ResourcesAvailableEntry']
_GCSNODEINFO = DESCRIPTOR.message_types_by_name['GcsNodeInfo']
_GCSNODEINFO_RESOURCESTOTALENTRY = _GCSNODEINFO.nested_types_by_name['ResourcesTotalEntry']
_JOBSAPIINFO = DESCRIPTOR.message_types_by_name['JobsAPIInfo']
_JOBSAPIINFO_METADATAENTRY = _JOBSAPIINFO.nested_types_by_name['MetadataEntry']
_JOBSAPIINFO_ENTRYPOINTRESOURCESENTRY = _JOBSAPIINFO.nested_types_by_name['EntrypointResourcesEntry']
_JOBTABLEDATA = DESCRIPTOR.message_types_by_name['JobTableData']
_WORKERTABLEDATA = DESCRIPTOR.message_types_by_name['WorkerTableData']
_WORKERTABLEDATA_WORKERINFOENTRY = _WORKERTABLEDATA.nested_types_by_name['WorkerInfoEntry']
_WORKERDELTADATA = DESCRIPTOR.message_types_by_name['WorkerDeltaData']
_RESOURCEMAP = DESCRIPTOR.message_types_by_name['ResourceMap']
_RESOURCEMAP_ITEMSENTRY = _RESOURCEMAP.nested_types_by_name['ItemsEntry']
_STOREDCONFIG = DESCRIPTOR.message_types_by_name['StoredConfig']
_NODERESOURCECHANGE = DESCRIPTOR.message_types_by_name['NodeResourceChange']
_NODERESOURCECHANGE_UPDATEDRESOURCESENTRY = _NODERESOURCECHANGE.nested_types_by_name['UpdatedResourcesEntry']
_PUBSUBMESSAGE = DESCRIPTOR.message_types_by_name['PubSubMessage']
_RESOURCEUPDATE = DESCRIPTOR.message_types_by_name['ResourceUpdate']
_RESOURCEUSAGEBROADCASTDATA = DESCRIPTOR.message_types_by_name['ResourceUsageBroadcastData']
_RESOURCEDEMAND = DESCRIPTOR.message_types_by_name['ResourceDemand']
_RESOURCEDEMAND_SHAPEENTRY = _RESOURCEDEMAND.nested_types_by_name['ShapeEntry']
_RESOURCELOAD = DESCRIPTOR.message_types_by_name['ResourceLoad']
_RESOURCESDATA = DESCRIPTOR.message_types_by_name['ResourcesData']
_RESOURCESDATA_RESOURCESAVAILABLEENTRY = _RESOURCESDATA.nested_types_by_name['ResourcesAvailableEntry']
_RESOURCESDATA_RESOURCESTOTALENTRY = _RESOURCESDATA.nested_types_by_name['ResourcesTotalEntry']
_RESOURCESDATA_RESOURCELOADENTRY = _RESOURCESDATA.nested_types_by_name['ResourceLoadEntry']
_RESOURCESDATA_RESOURCESNORMALTASKENTRY = _RESOURCESDATA.nested_types_by_name['ResourcesNormalTaskEntry']
_RESOURCEUSAGEBATCHDATA = DESCRIPTOR.message_types_by_name['ResourceUsageBatchData']
_PLACEMENTGROUPLOAD = DESCRIPTOR.message_types_by_name['PlacementGroupLoad']
_PLACEMENTGROUPSTATS = DESCRIPTOR.message_types_by_name['PlacementGroupStats']
_PLACEMENTGROUPTABLEDATA = DESCRIPTOR.message_types_by_name['PlacementGroupTableData']
_ACTORTABLEDATA_ACTORSTATE = _ACTORTABLEDATA.enum_types_by_name['ActorState']
_GCSNODEINFO_GCSNODESTATE = _GCSNODEINFO.enum_types_by_name['GcsNodeState']
_PLACEMENTGROUPSTATS_SCHEDULINGSTATE = _PLACEMENTGROUPSTATS.enum_types_by_name['SchedulingState']
_PLACEMENTGROUPTABLEDATA_PLACEMENTGROUPSTATE = _PLACEMENTGROUPTABLEDATA.enum_types_by_name['PlacementGroupState']
GcsEntry = _reflection.GeneratedProtocolMessageType('GcsEntry', (_message.Message,), {
  'DESCRIPTOR' : _GCSENTRY,
  '__module__' : 'src.ray.protobuf.gcs_pb2'
  # @@protoc_insertion_point(class_scope:ray.rpc.GcsEntry)
  })
_sym_db.RegisterMessage(GcsEntry)

ObjectTableData = _reflection.GeneratedProtocolMessageType('ObjectTableData', (_message.Message,), {
  'DESCRIPTOR' : _OBJECTTABLEDATA,
  '__module__' : 'src.ray.protobuf.gcs_pb2'
  # @@protoc_insertion_point(class_scope:ray.rpc.ObjectTableData)
  })
_sym_db.RegisterMessage(ObjectTableData)

ActorTableData = _reflection.GeneratedProtocolMessageType('ActorTableData', (_message.Message,), {

  'RequiredResourcesEntry' : _reflection.GeneratedProtocolMessageType('RequiredResourcesEntry', (_message.Message,), {
    'DESCRIPTOR' : _ACTORTABLEDATA_REQUIREDRESOURCESENTRY,
    '__module__' : 'src.ray.protobuf.gcs_pb2'
    # @@protoc_insertion_point(class_scope:ray.rpc.ActorTableData.RequiredResourcesEntry)
    })
  ,
  'DESCRIPTOR' : _ACTORTABLEDATA,
  '__module__' : 'src.ray.protobuf.gcs_pb2'
  # @@protoc_insertion_point(class_scope:ray.rpc.ActorTableData)
  })
_sym_db.RegisterMessage(ActorTableData)
_sym_db.RegisterMessage(ActorTableData.RequiredResourcesEntry)

ErrorTableData = _reflection.GeneratedProtocolMessageType('ErrorTableData', (_message.Message,), {
  'DESCRIPTOR' : _ERRORTABLEDATA,
  '__module__' : 'src.ray.protobuf.gcs_pb2'
  # @@protoc_insertion_point(class_scope:ray.rpc.ErrorTableData)
  })
_sym_db.RegisterMessage(ErrorTableData)

ScheduleData = _reflection.GeneratedProtocolMessageType('ScheduleData', (_message.Message,), {

  'SchedulePlanEntry' : _reflection.GeneratedProtocolMessageType('SchedulePlanEntry', (_message.Message,), {
    'DESCRIPTOR' : _SCHEDULEDATA_SCHEDULEPLANENTRY,
    '__module__' : 'src.ray.protobuf.gcs_pb2'
    # @@protoc_insertion_point(class_scope:ray.rpc.ScheduleData.SchedulePlanEntry)
    })
  ,
  'DESCRIPTOR' : _SCHEDULEDATA,
  '__module__' : 'src.ray.protobuf.gcs_pb2'
  # @@protoc_insertion_point(class_scope:ray.rpc.ScheduleData)
  })
_sym_db.RegisterMessage(ScheduleData)
_sym_db.RegisterMessage(ScheduleData.SchedulePlanEntry)

ProfileEventEntry = _reflection.GeneratedProtocolMessageType('ProfileEventEntry', (_message.Message,), {
  'DESCRIPTOR' : _PROFILEEVENTENTRY,
  '__module__' : 'src.ray.protobuf.gcs_pb2'
  # @@protoc_insertion_point(class_scope:ray.rpc.ProfileEventEntry)
  })
_sym_db.RegisterMessage(ProfileEventEntry)

ProfileEvents = _reflection.GeneratedProtocolMessageType('ProfileEvents', (_message.Message,), {
  'DESCRIPTOR' : _PROFILEEVENTS,
  '__module__' : 'src.ray.protobuf.gcs_pb2'
  # @@protoc_insertion_point(class_scope:ray.rpc.ProfileEvents)
  })
_sym_db.RegisterMessage(ProfileEvents)

TaskStateUpdate = _reflection.GeneratedProtocolMessageType('TaskStateUpdate', (_message.Message,), {
  'DESCRIPTOR' : _TASKSTATEUPDATE,
  '__module__' : 'src.ray.protobuf.gcs_pb2'
  # @@protoc_insertion_point(class_scope:ray.rpc.TaskStateUpdate)
  })
_sym_db.RegisterMessage(TaskStateUpdate)

TaskEvents = _reflection.GeneratedProtocolMessageType('TaskEvents', (_message.Message,), {
  'DESCRIPTOR' : _TASKEVENTS,
  '__module__' : 'src.ray.protobuf.gcs_pb2'
  # @@protoc_insertion_point(class_scope:ray.rpc.TaskEvents)
  })
_sym_db.RegisterMessage(TaskEvents)

TaskEventData = _reflection.GeneratedProtocolMessageType('TaskEventData', (_message.Message,), {
  'DESCRIPTOR' : _TASKEVENTDATA,
  '__module__' : 'src.ray.protobuf.gcs_pb2'
  # @@protoc_insertion_point(class_scope:ray.rpc.TaskEventData)
  })
_sym_db.RegisterMessage(TaskEventData)

ResourceTableData = _reflection.GeneratedProtocolMessageType('ResourceTableData', (_message.Message,), {
  'DESCRIPTOR' : _RESOURCETABLEDATA,
  '__module__' : 'src.ray.protobuf.gcs_pb2'
  # @@protoc_insertion_point(class_scope:ray.rpc.ResourceTableData)
  })
_sym_db.RegisterMessage(ResourceTableData)

AvailableResources = _reflection.GeneratedProtocolMessageType('AvailableResources', (_message.Message,), {

  'ResourcesAvailableEntry' : _reflection.GeneratedProtocolMessageType('ResourcesAvailableEntry', (_message.Message,), {
    'DESCRIPTOR' : _AVAILABLERESOURCES_RESOURCESAVAILABLEENTRY,
    '__module__' : 'src.ray.protobuf.gcs_pb2'
    # @@protoc_insertion_point(class_scope:ray.rpc.AvailableResources.ResourcesAvailableEntry)
    })
  ,
  'DESCRIPTOR' : _AVAILABLERESOURCES,
  '__module__' : 'src.ray.protobuf.gcs_pb2'
  # @@protoc_insertion_point(class_scope:ray.rpc.AvailableResources)
  })
_sym_db.RegisterMessage(AvailableResources)
_sym_db.RegisterMessage(AvailableResources.ResourcesAvailableEntry)

GcsNodeInfo = _reflection.GeneratedProtocolMessageType('GcsNodeInfo', (_message.Message,), {

  'ResourcesTotalEntry' : _reflection.GeneratedProtocolMessageType('ResourcesTotalEntry', (_message.Message,), {
    'DESCRIPTOR' : _GCSNODEINFO_RESOURCESTOTALENTRY,
    '__module__' : 'src.ray.protobuf.gcs_pb2'
    # @@protoc_insertion_point(class_scope:ray.rpc.GcsNodeInfo.ResourcesTotalEntry)
    })
  ,
  'DESCRIPTOR' : _GCSNODEINFO,
  '__module__' : 'src.ray.protobuf.gcs_pb2'
  # @@protoc_insertion_point(class_scope:ray.rpc.GcsNodeInfo)
  })
_sym_db.RegisterMessage(GcsNodeInfo)
_sym_db.RegisterMessage(GcsNodeInfo.ResourcesTotalEntry)

JobsAPIInfo = _reflection.GeneratedProtocolMessageType('JobsAPIInfo', (_message.Message,), {

  'MetadataEntry' : _reflection.GeneratedProtocolMessageType('MetadataEntry', (_message.Message,), {
    'DESCRIPTOR' : _JOBSAPIINFO_METADATAENTRY,
    '__module__' : 'src.ray.protobuf.gcs_pb2'
    # @@protoc_insertion_point(class_scope:ray.rpc.JobsAPIInfo.MetadataEntry)
    })
  ,

  'EntrypointResourcesEntry' : _reflection.GeneratedProtocolMessageType('EntrypointResourcesEntry', (_message.Message,), {
    'DESCRIPTOR' : _JOBSAPIINFO_ENTRYPOINTRESOURCESENTRY,
    '__module__' : 'src.ray.protobuf.gcs_pb2'
    # @@protoc_insertion_point(class_scope:ray.rpc.JobsAPIInfo.EntrypointResourcesEntry)
    })
  ,
  'DESCRIPTOR' : _JOBSAPIINFO,
  '__module__' : 'src.ray.protobuf.gcs_pb2'
  # @@protoc_insertion_point(class_scope:ray.rpc.JobsAPIInfo)
  })
_sym_db.RegisterMessage(JobsAPIInfo)
_sym_db.RegisterMessage(JobsAPIInfo.MetadataEntry)
_sym_db.RegisterMessage(JobsAPIInfo.EntrypointResourcesEntry)

JobTableData = _reflection.GeneratedProtocolMessageType('JobTableData', (_message.Message,), {
  'DESCRIPTOR' : _JOBTABLEDATA,
  '__module__' : 'src.ray.protobuf.gcs_pb2'
  # @@protoc_insertion_point(class_scope:ray.rpc.JobTableData)
  })
_sym_db.RegisterMessage(JobTableData)

WorkerTableData = _reflection.GeneratedProtocolMessageType('WorkerTableData', (_message.Message,), {

  'WorkerInfoEntry' : _reflection.GeneratedProtocolMessageType('WorkerInfoEntry', (_message.Message,), {
    'DESCRIPTOR' : _WORKERTABLEDATA_WORKERINFOENTRY,
    '__module__' : 'src.ray.protobuf.gcs_pb2'
    # @@protoc_insertion_point(class_scope:ray.rpc.WorkerTableData.WorkerInfoEntry)
    })
  ,
  'DESCRIPTOR' : _WORKERTABLEDATA,
  '__module__' : 'src.ray.protobuf.gcs_pb2'
  # @@protoc_insertion_point(class_scope:ray.rpc.WorkerTableData)
  })
_sym_db.RegisterMessage(WorkerTableData)
_sym_db.RegisterMessage(WorkerTableData.WorkerInfoEntry)

WorkerDeltaData = _reflection.GeneratedProtocolMessageType('WorkerDeltaData', (_message.Message,), {
  'DESCRIPTOR' : _WORKERDELTADATA,
  '__module__' : 'src.ray.protobuf.gcs_pb2'
  # @@protoc_insertion_point(class_scope:ray.rpc.WorkerDeltaData)
  })
_sym_db.RegisterMessage(WorkerDeltaData)

ResourceMap = _reflection.GeneratedProtocolMessageType('ResourceMap', (_message.Message,), {

  'ItemsEntry' : _reflection.GeneratedProtocolMessageType('ItemsEntry', (_message.Message,), {
    'DESCRIPTOR' : _RESOURCEMAP_ITEMSENTRY,
    '__module__' : 'src.ray.protobuf.gcs_pb2'
    # @@protoc_insertion_point(class_scope:ray.rpc.ResourceMap.ItemsEntry)
    })
  ,
  'DESCRIPTOR' : _RESOURCEMAP,
  '__module__' : 'src.ray.protobuf.gcs_pb2'
  # @@protoc_insertion_point(class_scope:ray.rpc.ResourceMap)
  })
_sym_db.RegisterMessage(ResourceMap)
_sym_db.RegisterMessage(ResourceMap.ItemsEntry)

StoredConfig = _reflection.GeneratedProtocolMessageType('StoredConfig', (_message.Message,), {
  'DESCRIPTOR' : _STOREDCONFIG,
  '__module__' : 'src.ray.protobuf.gcs_pb2'
  # @@protoc_insertion_point(class_scope:ray.rpc.StoredConfig)
  })
_sym_db.RegisterMessage(StoredConfig)

NodeResourceChange = _reflection.GeneratedProtocolMessageType('NodeResourceChange', (_message.Message,), {

  'UpdatedResourcesEntry' : _reflection.GeneratedProtocolMessageType('UpdatedResourcesEntry', (_message.Message,), {
    'DESCRIPTOR' : _NODERESOURCECHANGE_UPDATEDRESOURCESENTRY,
    '__module__' : 'src.ray.protobuf.gcs_pb2'
    # @@protoc_insertion_point(class_scope:ray.rpc.NodeResourceChange.UpdatedResourcesEntry)
    })
  ,
  'DESCRIPTOR' : _NODERESOURCECHANGE,
  '__module__' : 'src.ray.protobuf.gcs_pb2'
  # @@protoc_insertion_point(class_scope:ray.rpc.NodeResourceChange)
  })
_sym_db.RegisterMessage(NodeResourceChange)
_sym_db.RegisterMessage(NodeResourceChange.UpdatedResourcesEntry)

PubSubMessage = _reflection.GeneratedProtocolMessageType('PubSubMessage', (_message.Message,), {
  'DESCRIPTOR' : _PUBSUBMESSAGE,
  '__module__' : 'src.ray.protobuf.gcs_pb2'
  # @@protoc_insertion_point(class_scope:ray.rpc.PubSubMessage)
  })
_sym_db.RegisterMessage(PubSubMessage)

ResourceUpdate = _reflection.GeneratedProtocolMessageType('ResourceUpdate', (_message.Message,), {
  'DESCRIPTOR' : _RESOURCEUPDATE,
  '__module__' : 'src.ray.protobuf.gcs_pb2'
  # @@protoc_insertion_point(class_scope:ray.rpc.ResourceUpdate)
  })
_sym_db.RegisterMessage(ResourceUpdate)

ResourceUsageBroadcastData = _reflection.GeneratedProtocolMessageType('ResourceUsageBroadcastData', (_message.Message,), {
  'DESCRIPTOR' : _RESOURCEUSAGEBROADCASTDATA,
  '__module__' : 'src.ray.protobuf.gcs_pb2'
  # @@protoc_insertion_point(class_scope:ray.rpc.ResourceUsageBroadcastData)
  })
_sym_db.RegisterMessage(ResourceUsageBroadcastData)

ResourceDemand = _reflection.GeneratedProtocolMessageType('ResourceDemand', (_message.Message,), {

  'ShapeEntry' : _reflection.GeneratedProtocolMessageType('ShapeEntry', (_message.Message,), {
    'DESCRIPTOR' : _RESOURCEDEMAND_SHAPEENTRY,
    '__module__' : 'src.ray.protobuf.gcs_pb2'
    # @@protoc_insertion_point(class_scope:ray.rpc.ResourceDemand.ShapeEntry)
    })
  ,
  'DESCRIPTOR' : _RESOURCEDEMAND,
  '__module__' : 'src.ray.protobuf.gcs_pb2'
  # @@protoc_insertion_point(class_scope:ray.rpc.ResourceDemand)
  })
_sym_db.RegisterMessage(ResourceDemand)
_sym_db.RegisterMessage(ResourceDemand.ShapeEntry)

ResourceLoad = _reflection.GeneratedProtocolMessageType('ResourceLoad', (_message.Message,), {
  'DESCRIPTOR' : _RESOURCELOAD,
  '__module__' : 'src.ray.protobuf.gcs_pb2'
  # @@protoc_insertion_point(class_scope:ray.rpc.ResourceLoad)
  })
_sym_db.RegisterMessage(ResourceLoad)

ResourcesData = _reflection.GeneratedProtocolMessageType('ResourcesData', (_message.Message,), {

  'ResourcesAvailableEntry' : _reflection.GeneratedProtocolMessageType('ResourcesAvailableEntry', (_message.Message,), {
    'DESCRIPTOR' : _RESOURCESDATA_RESOURCESAVAILABLEENTRY,
    '__module__' : 'src.ray.protobuf.gcs_pb2'
    # @@protoc_insertion_point(class_scope:ray.rpc.ResourcesData.ResourcesAvailableEntry)
    })
  ,

  'ResourcesTotalEntry' : _reflection.GeneratedProtocolMessageType('ResourcesTotalEntry', (_message.Message,), {
    'DESCRIPTOR' : _RESOURCESDATA_RESOURCESTOTALENTRY,
    '__module__' : 'src.ray.protobuf.gcs_pb2'
    # @@protoc_insertion_point(class_scope:ray.rpc.ResourcesData.ResourcesTotalEntry)
    })
  ,

  'ResourceLoadEntry' : _reflection.GeneratedProtocolMessageType('ResourceLoadEntry', (_message.Message,), {
    'DESCRIPTOR' : _RESOURCESDATA_RESOURCELOADENTRY,
    '__module__' : 'src.ray.protobuf.gcs_pb2'
    # @@protoc_insertion_point(class_scope:ray.rpc.ResourcesData.ResourceLoadEntry)
    })
  ,

  'ResourcesNormalTaskEntry' : _reflection.GeneratedProtocolMessageType('ResourcesNormalTaskEntry', (_message.Message,), {
    'DESCRIPTOR' : _RESOURCESDATA_RESOURCESNORMALTASKENTRY,
    '__module__' : 'src.ray.protobuf.gcs_pb2'
    # @@protoc_insertion_point(class_scope:ray.rpc.ResourcesData.ResourcesNormalTaskEntry)
    })
  ,
  'DESCRIPTOR' : _RESOURCESDATA,
  '__module__' : 'src.ray.protobuf.gcs_pb2'
  # @@protoc_insertion_point(class_scope:ray.rpc.ResourcesData)
  })
_sym_db.RegisterMessage(ResourcesData)
_sym_db.RegisterMessage(ResourcesData.ResourcesAvailableEntry)
_sym_db.RegisterMessage(ResourcesData.ResourcesTotalEntry)
_sym_db.RegisterMessage(ResourcesData.ResourceLoadEntry)
_sym_db.RegisterMessage(ResourcesData.ResourcesNormalTaskEntry)

ResourceUsageBatchData = _reflection.GeneratedProtocolMessageType('ResourceUsageBatchData', (_message.Message,), {
  'DESCRIPTOR' : _RESOURCEUSAGEBATCHDATA,
  '__module__' : 'src.ray.protobuf.gcs_pb2'
  # @@protoc_insertion_point(class_scope:ray.rpc.ResourceUsageBatchData)
  })
_sym_db.RegisterMessage(ResourceUsageBatchData)

PlacementGroupLoad = _reflection.GeneratedProtocolMessageType('PlacementGroupLoad', (_message.Message,), {
  'DESCRIPTOR' : _PLACEMENTGROUPLOAD,
  '__module__' : 'src.ray.protobuf.gcs_pb2'
  # @@protoc_insertion_point(class_scope:ray.rpc.PlacementGroupLoad)
  })
_sym_db.RegisterMessage(PlacementGroupLoad)

PlacementGroupStats = _reflection.GeneratedProtocolMessageType('PlacementGroupStats', (_message.Message,), {
  'DESCRIPTOR' : _PLACEMENTGROUPSTATS,
  '__module__' : 'src.ray.protobuf.gcs_pb2'
  # @@protoc_insertion_point(class_scope:ray.rpc.PlacementGroupStats)
  })
_sym_db.RegisterMessage(PlacementGroupStats)

PlacementGroupTableData = _reflection.GeneratedProtocolMessageType('PlacementGroupTableData', (_message.Message,), {
  'DESCRIPTOR' : _PLACEMENTGROUPTABLEDATA,
  '__module__' : 'src.ray.protobuf.gcs_pb2'
  # @@protoc_insertion_point(class_scope:ray.rpc.PlacementGroupTableData)
  })
_sym_db.RegisterMessage(PlacementGroupTableData)

if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'\n\030io.ray.runtime.generated\370\001\001'
  _ACTORTABLEDATA_REQUIREDRESOURCESENTRY._options = None
  _ACTORTABLEDATA_REQUIREDRESOURCESENTRY._serialized_options = b'8\001'
  _SCHEDULEDATA_SCHEDULEPLANENTRY._options = None
  _SCHEDULEDATA_SCHEDULEPLANENTRY._serialized_options = b'8\001'
  _AVAILABLERESOURCES_RESOURCESAVAILABLEENTRY._options = None
  _AVAILABLERESOURCES_RESOURCESAVAILABLEENTRY._serialized_options = b'8\001'
  _GCSNODEINFO_RESOURCESTOTALENTRY._options = None
  _GCSNODEINFO_RESOURCESTOTALENTRY._serialized_options = b'8\001'
  _JOBSAPIINFO_METADATAENTRY._options = None
  _JOBSAPIINFO_METADATAENTRY._serialized_options = b'8\001'
  _JOBSAPIINFO_ENTRYPOINTRESOURCESENTRY._options = None
  _JOBSAPIINFO_ENTRYPOINTRESOURCESENTRY._serialized_options = b'8\001'
  _WORKERTABLEDATA_WORKERINFOENTRY._options = None
  _WORKERTABLEDATA_WORKERINFOENTRY._serialized_options = b'8\001'
  _RESOURCEMAP_ITEMSENTRY._options = None
  _RESOURCEMAP_ITEMSENTRY._serialized_options = b'8\001'
  _NODERESOURCECHANGE_UPDATEDRESOURCESENTRY._options = None
  _NODERESOURCECHANGE_UPDATEDRESOURCESENTRY._serialized_options = b'8\001'
  _RESOURCEDEMAND_SHAPEENTRY._options = None
  _RESOURCEDEMAND_SHAPEENTRY._serialized_options = b'8\001'
  _RESOURCESDATA_RESOURCESAVAILABLEENTRY._options = None
  _RESOURCESDATA_RESOURCESAVAILABLEENTRY._serialized_options = b'8\001'
  _RESOURCESDATA_RESOURCESTOTALENTRY._options = None
  _RESOURCESDATA_RESOURCESTOTALENTRY._serialized_options = b'8\001'
  _RESOURCESDATA_RESOURCELOADENTRY._options = None
  _RESOURCESDATA_RESOURCELOADENTRY._serialized_options = b'8\001'
  _RESOURCESDATA_RESOURCESNORMALTASKENTRY._options = None
  _RESOURCESDATA_RESOURCESNORMALTASKENTRY._serialized_options = b'8\001'
  _TABLEPREFIX._serialized_start=10221
  _TABLEPREFIX._serialized_end=10569
  _TABLEPUBSUB._serialized_start=10572
  _TABLEPUBSUB._serialized_end=10890
  _GCSCHANGEMODE._serialized_start=10892
  _GCSCHANGEMODE._serialized_end=10938
  _GCSENTRY._serialized_start=70
  _GCSENTRY._serialized_end=179
  _OBJECTTABLEDATA._serialized_start=181
  _OBJECTTABLEDATA._serialized_end=224
  _ACTORTABLEDATA._serialized_start=227
  _ACTORTABLEDATA._serialized_end=1485
  _ACTORTABLEDATA_REQUIREDRESOURCESENTRY._serialized_start=1283
  _ACTORTABLEDATA_REQUIREDRESOURCESENTRY._serialized_end=1351
  _ACTORTABLEDATA_ACTORSTATE._serialized_start=1353
  _ACTORTABLEDATA_ACTORSTATE._serialized_end=1450
  _ERRORTABLEDATA._serialized_start=1487
  _ERRORTABLEDATA._serialized_end=1613
  _SCHEDULEDATA._serialized_start=1616
  _SCHEDULEDATA._serialized_end=1773
  _SCHEDULEDATA_SCHEDULEPLANENTRY._serialized_start=1710
  _SCHEDULEDATA_SCHEDULEPLANENTRY._serialized_end=1773
  _PROFILEEVENTENTRY._serialized_start=1776
  _PROFILEEVENTENTRY._serialized_end=1935
  _PROFILEEVENTS._serialized_start=1938
  _PROFILEEVENTS._serialized_end=2119
  _TASKSTATEUPDATE._serialized_start=2122
  _TASKSTATEUPDATE._serialized_end=2646
  _TASKEVENTS._serialized_start=2649
  _TASKEVENTS._serialized_end=2993
  _TASKEVENTDATA._serialized_start=2996
  _TASKEVENTDATA._serialized_end=3208
  _RESOURCETABLEDATA._serialized_start=3210
  _RESOURCETABLEDATA._serialized_end=3274
  _AVAILABLERESOURCES._serialized_start=3277
  _AVAILABLERESOURCES._serialized_end=3495
  _AVAILABLERESOURCES_RESOURCESAVAILABLEENTRY._serialized_start=3426
  _AVAILABLERESOURCES_RESOURCESAVAILABLEENTRY._serialized_end=3495
  _GCSNODEINFO._serialized_start=3498
  _GCSNODEINFO._serialized_end=4222
  _GCSNODEINFO_RESOURCESTOTALENTRY._serialized_start=4120
  _GCSNODEINFO_RESOURCESTOTALENTRY._serialized_end=4185
  _GCSNODEINFO_GCSNODESTATE._serialized_start=4187
  _GCSNODEINFO_GCSNODESTATE._serialized_end=4222
  _JOBSAPIINFO._serialized_start=4225
  _JOBSAPIINFO._serialized_end=5157
  _JOBSAPIINFO_METADATAENTRY._serialized_start=4853
  _JOBSAPIINFO_METADATAENTRY._serialized_end=4912
  _JOBSAPIINFO_ENTRYPOINTRESOURCESENTRY._serialized_start=4914
  _JOBSAPIINFO_ENTRYPOINTRESOURCESENTRY._serialized_end=4984
  _JOBTABLEDATA._serialized_start=5160
  _JOBTABLEDATA._serialized_end=5528
  _WORKERTABLEDATA._serialized_start=5531
  _WORKERTABLEDATA._serialized_end=6146
  _WORKERTABLEDATA_WORKERINFOENTRY._serialized_start=6055
  _WORKERTABLEDATA_WORKERINFOENTRY._serialized_end=6116
  _WORKERDELTADATA._serialized_start=6148
  _WORKERDELTADATA._serialized_end=6223
  _RESOURCEMAP._serialized_start=6226
  _RESOURCEMAP._serialized_end=6380
  _RESOURCEMAP_ITEMSENTRY._serialized_start=6296
  _RESOURCEMAP_ITEMSENTRY._serialized_end=6380
  _STOREDCONFIG._serialized_start=6382
  _STOREDCONFIG._serialized_end=6420
  _NODERESOURCECHANGE._serialized_start=6423
  _NODERESOURCECHANGE._serialized_end=6678
  _NODERESOURCECHANGE_UPDATEDRESOURCESENTRY._serialized_start=6611
  _NODERESOURCECHANGE_UPDATEDRESOURCESENTRY._serialized_end=6678
  _PUBSUBMESSAGE._serialized_start=6680
  _PUBSUBMESSAGE._serialized_end=6731
  _RESOURCEUPDATE._serialized_start=6734
  _RESOURCEUPDATE._serialized_end=6878
  _RESOURCEUSAGEBROADCASTDATA._serialized_start=6880
  _RESOURCEUSAGEBROADCASTDATA._serialized_end=6978
  _RESOURCEDEMAND._serialized_start=6981
  _RESOURCEDEMAND._serialized_end=7276
  _RESOURCEDEMAND_SHAPEENTRY._serialized_start=7220
  _RESOURCEDEMAND_SHAPEENTRY._serialized_end=7276
  _RESOURCELOAD._serialized_start=7278
  _RESOURCELOAD._serialized_end=7360
  _RESOURCESDATA._serialized_start=7363
  _RESOURCESDATA._serialized_end=8580
  _RESOURCESDATA_RESOURCESAVAILABLEENTRY._serialized_start=3426
  _RESOURCESDATA_RESOURCESAVAILABLEENTRY._serialized_end=3495
  _RESOURCESDATA_RESOURCESTOTALENTRY._serialized_start=4120
  _RESOURCESDATA_RESOURCESTOTALENTRY._serialized_end=4185
  _RESOURCESDATA_RESOURCELOADENTRY._serialized_start=8445
  _RESOURCESDATA_RESOURCELOADENTRY._serialized_end=8508
  _RESOURCESDATA_RESOURCESNORMALTASKENTRY._serialized_start=8510
  _RESOURCESDATA_RESOURCESNORMALTASKENTRY._serialized_end=8580
  _RESOURCEUSAGEBATCHDATA._serialized_start=8583
  _RESOURCEUSAGEBATCHDATA._serialized_end=8808
  _PLACEMENTGROUPLOAD._serialized_start=8810
  _PLACEMENTGROUPLOAD._serialized_end=8914
  _PLACEMENTGROUPSTATS._serialized_start=8917
  _PLACEMENTGROUPSTATS._serialized_end=9521
  _PLACEMENTGROUPSTATS_SCHEDULINGSTATE._serialized_start=9375
  _PLACEMENTGROUPSTATS_SCHEDULINGSTATE._serialized_end=9521
  _PLACEMENTGROUPTABLEDATA._serialized_start=9524
  _PLACEMENTGROUPTABLEDATA._serialized_end=10218
  _PLACEMENTGROUPTABLEDATA_PLACEMENTGROUPSTATE._serialized_start=10140
  _PLACEMENTGROUPTABLEDATA_PLACEMENTGROUPSTATE._serialized_end=10218
# @@protoc_insertion_point(module_scope)



from dataclasses import dataclass, field
from typing import Dict, Any, Callable, Optional, Type, List


@dataclass
class Config:
    type_hooks: Dict[Type, Callable[[Any], Any]] = field(default_factory=dict)
    cast: List[Type] = field(default_factory=list)
    forward_references: Optional[Dict[str, Any]] = None
    check_types: bool = True
    strict: bool = False
    strict_unions_match: bool = False



from ray.rllib.algorithms.ars.ars import ARS as ARSTrainer, DEFAULT_CONFIG
from ray.rllib.algorithms.ars.ars_tf_policy import ARSTFPolicy
from ray.rllib.algorithms.ars.ars_torch_policy import ARSTorchPolicy

__all__ = [
    "ARSTFPolicy",
    "ARSTorchPolicy",
    "ARSTrainer",
    "DEFAULT_CONFIG",
]

from ray.rllib.utils.deprecation import deprecation_warning

deprecation_warning("ray.rllib.agents.ars", "ray.rllib.algorithms.ars", error=True)



import atexit
import logging
import random
import time
from functools import wraps
from typing import Any, Callable, Dict, Iterable, List, Optional, Tuple, Type, Union

import ray
from ray.actor import ActorHandle
from ray.serve._private.common import (
    DeploymentInfo,
    DeploymentStatus,
    StatusOverview,
    ApplicationStatus,
    DeploymentStatusInfo,
)
from ray.serve.config import DeploymentConfig, HTTPOptions, ReplicaConfig
from ray.serve._private.constants import (
    CLIENT_POLLING_INTERVAL_S,
    MAX_CACHED_HANDLES,
    SERVE_NAMESPACE,
)
from ray.serve.controller import ServeController
from ray.serve.exceptions import RayServeException
from ray.serve.generated.serve_pb2 import DeploymentRoute, DeploymentRouteList
from ray.serve.generated.serve_pb2 import StatusOverview as StatusOverviewProto
from ray.serve.generated.serve_pb2 import (
    DeploymentStatusInfo as DeploymentStatusInfoProto,
)
from ray.serve.handle import RayServeHandle, RayServeSyncHandle
from ray.serve.schema import ServeApplicationSchema

logger = logging.getLogger(__file__)


def _ensure_connected(f: Callable) -> Callable:
    @wraps(f)
    def check(self, *args, **kwargs):
        if self._shutdown:
            raise RayServeException("Client has already been shut down.")
        return f(self, *args, **kwargs)

    return check


class ServeControllerClient:
    def __init__(
        self,
        controller: ActorHandle,
        controller_name: str,
        detached: bool = False,
    ):
        self._controller: ServeController = controller
        self._controller_name = controller_name
        self._detached = detached
        self._shutdown = False
        self._http_config: HTTPOptions = ray.get(controller.get_http_config.remote())
        self._root_url = ray.get(controller.get_root_url.remote())

        # Each handle has the overhead of long poll client, therefore cached.
        self.handle_cache = dict()
        self._evicted_handle_keys = set()

        # NOTE(edoakes): Need this because the shutdown order isn't guaranteed
        # when the interpreter is exiting so we can't rely on __del__ (it
        # throws a nasty stacktrace).
        if not self._detached:

            def shutdown_serve_client():
                self.shutdown()

            atexit.register(shutdown_serve_client)

    @property
    def root_url(self):
        return self._root_url

    @property
    def http_config(self):
        return self._http_config

    def __del__(self):
        if not self._detached:
            logger.debug(
                "Shutting down Ray Serve because client went out of "
                "scope. To prevent this, either keep a reference to "
                "the client or use serve.start(detached=True)."
            )
            self.shutdown()

    def __reduce__(self):
        raise RayServeException(("Ray Serve client cannot be serialized."))

    def shutdown(self) -> None:
        """Completely shut down the connected Serve instance.

        Shuts down all processes and deletes all state associated with the
        instance.
        """

        # Shut down handles
        for k in list(self.handle_cache):
            self.handle_cache[k].stop_metrics_pusher()
            del self.handle_cache[k]

        if ray.is_initialized() and not self._shutdown:
            ray.get(self._controller.shutdown.remote())
            self._wait_for_deployments_shutdown()

            ray.kill(self._controller, no_restart=True)

            # Wait for the named actor entry gets removed as well.
            started = time.time()
            while True:
                try:
                    ray.get_actor(self._controller_name, namespace=SERVE_NAMESPACE)
                    if time.time() - started > 5:
                        logger.warning(
                            "Waited 5s for Serve to shutdown gracefully but "
                            "the controller is still not cleaned up. "
                            "You can ignore this warning if you are shutting "
                            "down the Ray cluster."
                        )
                        break
                except ValueError:  # actor name is removed
                    break

            self._shutdown = True

    def _wait_for_deployments_shutdown(self, timeout_s: int = 60):
        """Waits for all deployments to be shut down and deleted.

        Raises TimeoutError if this doesn't happen before timeout_s.
        """
        start = time.time()
        while time.time() - start < timeout_s:
            deployment_statuses = self.get_serve_status().deployment_statuses
            if len(deployment_statuses) == 0:
                break
            else:
                logger.debug(
                    f"Waiting for shutdown, {len(deployment_statuses)} "
                    "deployments still alive."
                )
            time.sleep(CLIENT_POLLING_INTERVAL_S)
        else:
            live_names = [
                deployment_status.name for deployment_status in deployment_statuses
            ]
            raise TimeoutError(
                f"Shutdown didn't complete after {timeout_s}s. "
                f"Deployments still alive: {live_names}."
            )

    def _wait_for_deployment_healthy(self, name: str, timeout_s: int = -1):
        """Waits for the named deployment to enter "HEALTHY" status.

        Raises RuntimeError if the deployment enters the "UNHEALTHY" status
        instead.

        Raises TimeoutError if this doesn't happen before timeout_s.
        """
        start = time.time()
        while time.time() - start < timeout_s or timeout_s < 0:

            status_bytes = ray.get(self._controller.get_deployment_status.remote(name))

            if status_bytes is None:
                raise RuntimeError(
                    f"Waiting for deployment {name} to be HEALTHY, "
                    "but deployment doesn't exist."
                )

            status = DeploymentStatusInfo.from_proto(
                DeploymentStatusInfoProto.FromString(status_bytes)
            )

            if status.status == DeploymentStatus.HEALTHY:
                break
            elif status.status == DeploymentStatus.UNHEALTHY:
                raise RuntimeError(
                    f"Deployment {name} is UNHEALTHY: " f"{status.message}"
                )
            else:
                # Guard against new unhandled statuses being added.
                assert status.status == DeploymentStatus.UPDATING

            logger.debug(
                f"Waiting for {name} to be healthy, current status: "
                f"{status.status}."
            )
            time.sleep(CLIENT_POLLING_INTERVAL_S)
        else:
            raise TimeoutError(
                f"Deployment {name} did not become HEALTHY after {timeout_s}s."
            )

    def _wait_for_deployment_deleted(self, name: str, timeout_s: int = 60):
        """Waits for the named deployment to be shut down and deleted.

        Raises TimeoutError if this doesn't happen before timeout_s.
        """
        start = time.time()
        while time.time() - start < timeout_s:
            curr_status_bytes = ray.get(
                self._controller.get_deployment_status.remote(name)
            )
            if curr_status_bytes is None:
                break
            curr_status = DeploymentStatusInfo.from_proto(
                DeploymentStatusInfoProto.FromString(curr_status_bytes)
            )
            logger.debug(
                f"Waiting for {name} to be deleted, current status: {curr_status}."
            )
            time.sleep(CLIENT_POLLING_INTERVAL_S)
        else:
            raise TimeoutError(f"Deployment {name} wasn't deleted after {timeout_s}s.")

    @_ensure_connected
    def deploy(
        self,
        name: str,
        deployment_def: Union[Callable, Type[Callable], str],
        init_args: Tuple[Any],
        init_kwargs: Dict[Any, Any],
        ray_actor_options: Optional[Dict] = None,
        config: Optional[Union[DeploymentConfig, Dict[str, Any]]] = None,
        version: Optional[str] = None,
        route_prefix: Optional[str] = None,
        url: Optional[str] = None,
        _blocking: Optional[bool] = True,
    ):

        controller_deploy_args = self.get_deploy_args(
            name=name,
            deployment_def=deployment_def,
            init_args=init_args,
            init_kwargs=init_kwargs,
            ray_actor_options=ray_actor_options,
            config=config,
            version=version,
            route_prefix=route_prefix,
        )

        updating = ray.get(self._controller.deploy.remote(**controller_deploy_args))

        tag = self.log_deployment_update_status(name, version, updating)

        if _blocking:
            self._wait_for_deployment_healthy(name)
            self.log_deployment_ready(name, version, url, tag)

    @_ensure_connected
    def deploy_group(
        self,
        name,
        deployments: List[Dict],
        _blocking: bool = True,
        remove_past_deployments: bool = True,
    ):
        deployment_args_list = []
        for deployment in deployments:
            deployment_args_list.append(
                self.get_deploy_args(
                    deployment["name"],
                    deployment["func_or_class"],
                    deployment["init_args"],
                    deployment["init_kwargs"],
                    ray_actor_options=deployment["ray_actor_options"],
                    config=deployment["config"],
                    version=deployment["version"],
                    route_prefix=deployment["route_prefix"],
                    is_driver_deployment=deployment["is_driver_deployment"],
                )
            )

        updating_list = ray.get(
            self._controller.deploy_group.remote(name, deployment_args_list)
        )

        tags = []
        for i, updating in enumerate(updating_list):
            deployment = deployments[i]
            name, version = deployment["name"], deployment["version"]

            tags.append(self.log_deployment_update_status(name, version, updating))

        for i, deployment in enumerate(deployments):
            name = deployment["name"]
            url = deployment["url"]

            if _blocking:
                self._wait_for_deployment_healthy(name)
                self.log_deployment_ready(name, version, url, tags[i])

        if remove_past_deployments:
            # clean up the old deployments
            new_deployments_names = set()
            for deployment in deployments:
                new_deployments_names.add(deployment["name"])

            all_deployments_names = set(self.list_deployments().keys())
            deployment_names_to_delete = all_deployments_names.difference(
                new_deployments_names
            )
            self.delete_deployments(deployment_names_to_delete, blocking=_blocking)

    @_ensure_connected
    def deploy_app(
        self, config: ServeApplicationSchema, _blocking: bool = False
    ) -> None:
        ray.get(self._controller.deploy_app.remote(config))

        if _blocking:
            timeout_s = 60

            start = time.time()
            while time.time() - start < timeout_s:
                curr_status = self.get_serve_status()
                if curr_status.app_status.status == ApplicationStatus.RUNNING:
                    break
                time.sleep(CLIENT_POLLING_INTERVAL_S)
            else:
                raise TimeoutError(
                    f"Serve application isn't running after {timeout_s}s."
                )

    @_ensure_connected
    def delete_apps(self, names: List[str], blocking: bool = True):
        logger.info(f"Deleting app {names}")
        self._controller.delete_apps.remote(names)
        if blocking:
            start = time.time()
            while time.time() - start < 60:
                curr_statuses_bytes = ray.get(
                    self._controller.get_serve_statuses.remote(names)
                )
                all_deleted = True
                for cur_status_bytes in curr_statuses_bytes:
                    cur_status = StatusOverview.from_proto(
                        StatusOverviewProto.FromString(cur_status_bytes)
                    )
                    if cur_status.app_status.status != ApplicationStatus.NOT_STARTED:
                        all_deleted = False
                if all_deleted:
                    return
                time.sleep(CLIENT_POLLING_INTERVAL_S)
            else:
                raise TimeoutError(f"Deployment {names} wasn't deleted after 60s.")

    @_ensure_connected
    def delete_deployments(self, names: Iterable[str], blocking: bool = True) -> None:
        ray.get(self._controller.delete_deployments.remote(names))
        if blocking:
            for name in names:
                self._wait_for_deployment_deleted(name)

    @_ensure_connected
    def get_deployment_info(self, name: str) -> Tuple[DeploymentInfo, str]:
        deployment_route = DeploymentRoute.FromString(
            ray.get(self._controller.get_deployment_info.remote(name))
        )
        return (
            DeploymentInfo.from_proto(deployment_route.deployment_info),
            deployment_route.route if deployment_route.route != "" else None,
        )

    @_ensure_connected
    def list_deployments(self) -> Dict[str, Tuple[DeploymentInfo, str]]:
        deployment_route_list = DeploymentRouteList.FromString(
            ray.get(self._controller.list_deployments.remote())
        )
        return {
            deployment_route.deployment_info.name: (
                DeploymentInfo.from_proto(deployment_route.deployment_info),
                deployment_route.route if deployment_route.route != "" else None,
            )
            for deployment_route in deployment_route_list.deployment_routes
        }

    @_ensure_connected
    def get_app_config(self) -> Dict:
        """Returns the most recently requested Serve config."""
        return ray.get(self._controller.get_app_config.remote())

    @_ensure_connected
    def get_serve_status(self, name: str = "") -> StatusOverview:
        proto = StatusOverviewProto.FromString(
            ray.get(self._controller.get_serve_status.remote(name))
        )
        return StatusOverview.from_proto(proto)

    @_ensure_connected
    def get_handle(
        self,
        deployment_name: str,
        missing_ok: Optional[bool] = False,
        sync: bool = True,
        _internal_pickled_http_request: bool = False,
    ) -> Union[RayServeHandle, RayServeSyncHandle]:
        """Retrieve RayServeHandle for service deployment to invoke it from Python.

        Args:
            deployment_name: A registered service deployment.
            missing_ok: If true, then Serve won't check the deployment
                is registered. False by default.
            sync: If true, then Serve will return a ServeHandle that
                works everywhere. Otherwise, Serve will return a ServeHandle
                that's only usable in asyncio loop.

        Returns:
            RayServeHandle
        """
        cache_key = (deployment_name, missing_ok, sync)
        if cache_key in self.handle_cache:
            cached_handle = self.handle_cache[cache_key]
            if cached_handle.is_polling and cached_handle.is_same_loop:
                return cached_handle

        all_endpoints = ray.get(self._controller.get_all_endpoints.remote())
        if not missing_ok and deployment_name not in all_endpoints:
            raise KeyError(f"Deployment '{deployment_name}' does not exist.")

        if sync:
            handle = RayServeSyncHandle(
                self._controller,
                deployment_name,
                _internal_pickled_http_request=_internal_pickled_http_request,
            )
        else:
            handle = RayServeHandle(
                self._controller,
                deployment_name,
                _internal_pickled_http_request=_internal_pickled_http_request,
            )

        self.handle_cache[cache_key] = handle
        if cache_key in self._evicted_handle_keys:
            logger.warning(
                "You just got a ServeHandle that was evicted from internal "
                "cache. This means you are getting too many ServeHandles in "
                "the same process, this will bring down Serve's performance. "
                "Please post a github issue at "
                "https://github.com/ray-project/ray/issues to let the Serve "
                "team to find workaround for your use case."
            )

        if len(self.handle_cache) > MAX_CACHED_HANDLES:
            # Perform random eviction to keep the handle cache from growing
            # infinitely. We used use WeakValueDictionary but hit
            # https://github.com/ray-project/ray/issues/18980.
            evict_key = random.choice(list(self.handle_cache.keys()))
            self._evicted_handle_keys.add(evict_key)
            self.handle_cache.pop(evict_key)

        return handle

    @_ensure_connected
    def get_deploy_args(
        self,
        name: str,
        deployment_def: Union[Callable, Type[Callable], str],
        init_args: Tuple[Any],
        init_kwargs: Dict[Any, Any],
        ray_actor_options: Optional[Dict] = None,
        config: Optional[Union[DeploymentConfig, Dict[str, Any]]] = None,
        version: Optional[str] = None,
        route_prefix: Optional[str] = None,
        is_driver_deployment: Optional[str] = None,
    ) -> Dict:
        """
        Takes a deployment's configuration, and returns the arguments needed
        for the controller to deploy it.
        """

        if config is None:
            config = {}
        if ray_actor_options is None:
            ray_actor_options = {}

        curr_job_env = ray.get_runtime_context().runtime_env
        if "runtime_env" in ray_actor_options:
            # It is illegal to set field working_dir to None.
            if curr_job_env.get("working_dir") is not None:
                ray_actor_options["runtime_env"].setdefault(
                    "working_dir", curr_job_env.get("working_dir")
                )
        else:
            ray_actor_options["runtime_env"] = curr_job_env

        replica_config = ReplicaConfig.create(
            deployment_def,
            init_args=init_args,
            init_kwargs=init_kwargs,
            ray_actor_options=ray_actor_options,
        )

        if isinstance(config, dict):
            deployment_config = DeploymentConfig.parse_obj(config)
        elif isinstance(config, DeploymentConfig):
            deployment_config = config
        else:
            raise TypeError("config must be a DeploymentConfig or a dictionary.")

        deployment_config.version = version

        if (
            deployment_config.autoscaling_config is not None
            and deployment_config.max_concurrent_queries
            < deployment_config.autoscaling_config.target_num_ongoing_requests_per_replica  # noqa: E501
        ):
            logger.warning(
                "Autoscaling will never happen, "
                "because 'max_concurrent_queries' is less than "
                "'target_num_ongoing_requests_per_replica' now."
            )

        controller_deploy_args = {
            "name": name,
            "deployment_config_proto_bytes": deployment_config.to_proto_bytes(),
            "replica_config_proto_bytes": replica_config.to_proto_bytes(),
            "route_prefix": route_prefix,
            "deployer_job_id": ray.get_runtime_context().get_job_id(),
            "is_driver_deployment": is_driver_deployment,
        }

        return controller_deploy_args

    @_ensure_connected
    def log_deployment_update_status(
        self, name: str, version: str, updating: bool
    ) -> str:
        tag = f"component=serve deployment={name}"

        if updating:
            msg = f"Updating deployment '{name}'"
            if version is not None:
                msg += f" to version '{version}'"
            logger.info(f"{msg}. {tag}")
        else:
            logger.info(
                f"Deployment '{name}' is already at version "
                f"'{version}', not updating. {tag}"
            )

        return tag

    @_ensure_connected
    def log_deployment_ready(self, name: str, version: str, url: str, tag: str) -> None:
        if url is not None:
            url_part = f" at `{url}`"
        else:
            url_part = ""
        logger.info(
            f"Deployment '{name}{':'+version if version else ''}' is ready"
            f"{url_part}. {tag}"
        )



import json
import os
import shutil
import time
import unittest
from pathlib import Path
import logging

import pytest

import ray
from ray import tune
from ray.air import (
    Checkpoint,
    CheckpointConfig,
    FailureConfig,
    RunConfig,
    ScalingConfig,
    session,
)
from ray.air._internal.remote_storage import (
    delete_at_uri,
    download_from_uri,
    upload_to_uri,
    list_at_uri,
)
from ray.train.data_parallel_trainer import DataParallelTrainer
from ray.tune import Callback, Trainable
from ray.tune.execution.trial_runner import _find_newest_experiment_checkpoint
from ray.tune.experiment import Trial
from ray.tune.result_grid import ResultGrid
from ray.tune.schedulers.async_hyperband import ASHAScheduler
from ray.tune.search.optuna import OptunaSearch
from ray.tune.tune_config import TuneConfig
from ray.tune.tuner import Tuner


@pytest.fixture
def ray_start_2_cpus():
    address_info = ray.init(num_cpus=2, configure_logging=False)
    yield address_info
    # The code after the yield will run as teardown code.
    ray.shutdown()


@pytest.fixture
def ray_start_4_cpus():
    address_info = ray.init(num_cpus=4, configure_logging=False)
    yield address_info
    # The code after the yield will run as teardown code.
    ray.shutdown()


@pytest.fixture
def chdir_tmpdir(tmpdir):
    old_cwd = os.getcwd()
    os.chdir(tmpdir)
    yield tmpdir
    os.chdir(old_cwd)


@pytest.fixture
def clear_memory_filesys():
    yield
    delete_at_uri("memory:///")


def _train_fn_sometimes_failing(config):
    # Fails if failing is set and marker file exists.
    # Hangs if hanging is set and marker file exists.
    failing, hanging = config["failing_hanging"]

    checkpoint = session.get_checkpoint()
    if checkpoint:
        checkpoint_dict = checkpoint.to_dict()
        state = {"it": checkpoint_dict["it"]}
    else:
        state = {"it": 0}

    for i in range(config.get("num_epochs", 1)):
        state["it"] += 1

        session.report(state, checkpoint=Checkpoint.from_dict(state))

    # We fail after reporting num_epochs checkpoints.
    if failing and failing.exists():
        raise RuntimeError("I am failing")

    if hanging and hanging.exists():
        time.sleep(60)

    state["it"] += 1
    session.report(state, checkpoint=Checkpoint.from_dict(state))


class _FailOnStats(Callback):
    """Fail when at least num_trials exist and num_finished have finished."""

    def __init__(self, num_trials: int, num_finished: int, delay: int = 1):
        self.num_trials = num_trials
        self.num_finished = num_finished
        self.delay = delay
        self.fail_at = None

    def on_step_begin(self, iteration: int, trials: list, **info):
        if self.fail_at and iteration >= self.fail_at:
            print(
                "Actually failing after delay:",
                [(t.status, t.last_result.get("it")) for t in trials],
            )
            raise RuntimeError("Failing")

        if len(trials) < self.num_trials:
            return

        if (
            len([t for t in trials if t.status in [Trial.TERMINATED, Trial.ERROR]])
            >= self.num_finished
        ):
            self.fail_at = iteration + self.delay
            print(
                f"Triggering fail in {self.delay} iterations:",
                [(t.status, t.last_result.get("it")) for t in trials],
            )
        else:
            print("Not failing:", [(t.status, t.last_result.get("it")) for t in trials])


class MockData:
    def __init__(self):
        import numpy as np

        self.data = np.random.rand((2 * 1024 * 1024))


def test_tuner_restore_num_trials(ray_start_4_cpus, tmpdir):
    """Number of trials after restoring a finished run should be the same"""
    tuner = Tuner(
        lambda config: 1,
        tune_config=TuneConfig(num_samples=4, metric="_metric", mode="max"),
        run_config=RunConfig(
            name="test_tuner_restore_num_trials", local_dir=str(tmpdir)
        ),
    )
    results = tuner.fit()
    assert len(results) == 4
    assert results.get_best_result().metrics["_metric"] == 1

    del tuner
    tuner = Tuner.restore(str(tmpdir / "test_tuner_restore_num_trials"))

    # Check restored results
    results = tuner.get_results()
    assert len(results) == 4
    assert results.get_best_result().metrics["_metric"] == 1

    results = tuner.fit()
    assert len(results) == 4
    assert results.get_best_result().metrics["_metric"] == 1


def test_tuner_restore_resume_errored(ray_start_4_cpus, tmpdir):
    """Resuming errored trials should pick up from previous state"""
    fail_marker = tmpdir / "fail_marker"
    fail_marker.write_text("", encoding="utf-8")

    tuner = Tuner(
        _train_fn_sometimes_failing,
        tune_config=TuneConfig(
            num_samples=1,
        ),
        run_config=RunConfig(
            name="test_tuner_restore_resume_errored", local_dir=str(tmpdir)
        ),
        param_space={
            # Second and third trial fail
            "failing_hanging": tune.grid_search(
                [(None, None), (fail_marker, None), (None, None), (fail_marker, None)]
            ),
        },
    )
    results = tuner.fit()

    assert len(results) == 4
    assert len(results.errors) == 2
    # Second and third trial are at iter 1 because they failed after first report
    assert [r.metrics["it"] for r in results] == [2, 1, 2, 1]

    del tuner
    fail_marker.remove(ignore_errors=True)

    tuner = Tuner.restore(
        str(tmpdir / "test_tuner_restore_resume_errored"), resume_errored=True
    )

    # Check restored results
    results = tuner.get_results()
    assert len(results) == 4
    assert len(results.errors) == 2
    # Second and third trial are at iter 1 because they failed after first report
    assert [r.metrics["it"] for r in results] == [2, 1, 2, 1]

    # Get new results
    results = tuner.fit()

    assert len(results) == 4
    assert len(results.errors) == 0
    # Since the errored trials are being resumed from previous state and then report
    # two more times, we should observe 3 here.
    assert sorted([r.metrics["it"] for r in results]) == sorted([2, 3, 2, 3])


def test_tuner_restore_restart_errored(ray_start_4_cpus, tmpdir):
    """Restarting errored trials should re-start from scratch"""
    fail_marker = tmpdir / "fail_marker"
    fail_marker.write_text("", encoding="utf-8")

    tuner = Tuner(
        _train_fn_sometimes_failing,
        tune_config=TuneConfig(num_samples=1),
        run_config=RunConfig(
            name="test_tuner_restore_restart_errored",
            local_dir=str(tmpdir),
        ),
        param_space={
            # Second and third trial fail
            "failing_hanging": tune.grid_search(
                [(None, None), (fail_marker, None), (None, None), (fail_marker, None)]
            ),
        },
    )
    results = tuner.fit()

    assert len(results) == 4
    assert len(results.errors) == 2
    assert [r.metrics["it"] for r in results] == [2, 1, 2, 1]

    del tuner
    fail_marker.remove(ignore_errors=True)

    tuner = Tuner.restore(
        str(tmpdir / "test_tuner_restore_restart_errored"), restart_errored=True
    )

    # Check restored results
    results = tuner.get_results()
    assert len(results) == 4
    assert len(results.errors) == 2
    assert [r.metrics["it"] for r in results] == [2, 1, 2, 1]

    # Get new results
    results = tuner.fit()
    assert len(results) == 4
    assert len(results.errors) == 0
    # Since the errored trials are being restarted from scratch, they should report 2
    assert [r.metrics["it"] for r in results] == [2, 2, 2, 2]


def test_tuner_resume_unfinished(ray_start_2_cpus, tmpdir):
    """Resuming unfinished trials should pick up existing state"""
    os.environ["TUNE_GLOBAL_CHECKPOINT_S"] = "0"

    fail_marker = tmpdir / "fail_marker"
    fail_marker.write_text("", encoding="utf-8")

    hang_marker = tmpdir / "hang_marker"
    hang_marker.write_text("", encoding="utf-8")

    tuner = Tuner(
        _train_fn_sometimes_failing,
        tune_config=TuneConfig(num_samples=1),
        run_config=RunConfig(
            name="test_tuner_resume_unfinished",
            local_dir=str(tmpdir),
            failure_config=FailureConfig(fail_fast=False),
            callbacks=[_FailOnStats(num_trials=4, num_finished=2, delay=1)],
        ),
        param_space={
            # First trial succeeds, second hangs, third fails, fourth hangs
            "failing_hanging": tune.grid_search(
                [
                    (None, None),
                    (None, hang_marker),
                    (fail_marker, None),
                    (None, hang_marker),
                ]
            ),
        },
    )
    # Catch the FailOnStats error
    with pytest.raises(RuntimeError):
        tuner.fit()

    # After this run we have the following trial states (status, metric):
    # [('TERMINATED', 2), ('RUNNING', 1), ('ERROR', 1), ('PENDING', None)]

    # Restarting without hanging/failing should lead to the results:
    # [2, 3, 1, 2], because:
    # the TERMINATED trial is finished (state = 2),
    # the RUNNING trial is continued (and picks up from state = 1 for 2 iterations),
    # the ERROR trial is not continued (remains at 1 and errored)
    # and the PENDING trial has not state, yet.

    del tuner
    fail_marker.remove(ignore_errors=True)
    hang_marker.remove(ignore_errors=True)

    tuner = Tuner.restore(str(tmpdir / "test_tuner_resume_unfinished"))
    tuner._local_tuner._run_config.callbacks = None

    results = tuner.fit()
    assert len(results) == 4
    assert len(results.errors) == 1
    assert sorted([r.metrics["it"] for r in results]) == sorted([2, 3, 1, 2])


def test_tuner_resume_errored_only(ray_start_2_cpus, tmpdir):
    """Not resuming unfinished trials (but only errored and pending) should work"""
    os.environ["TUNE_GLOBAL_CHECKPOINT_S"] = "0"

    fail_marker = tmpdir / "fail_marker"
    fail_marker.write_text("", encoding="utf-8")

    hang_marker = tmpdir / "hang_marker"
    hang_marker.write_text("", encoding="utf-8")

    tuner = Tuner(
        _train_fn_sometimes_failing,
        tune_config=TuneConfig(num_samples=1),
        run_config=RunConfig(
            name="test_tuner_resume_errored_only",
            local_dir=str(tmpdir),
            failure_config=FailureConfig(fail_fast=False),
            callbacks=[_FailOnStats(num_trials=4, num_finished=2, delay=1)],
        ),
        param_space={
            # First trial succeeds, second hangs, third fails, fourth hangs
            "failing_hanging": tune.grid_search(
                [
                    (None, None),
                    (None, hang_marker),
                    (fail_marker, None),
                    (None, hang_marker),
                ]
            ),
        },
    )
    # Catch the FailOnStats error
    with pytest.raises(RuntimeError):
        tuner.fit()

    # After this run we have the following trial states (status, metric):
    # [('TERMINATED', 2), ('RUNNING', 1), ('ERROR', 1), ('PENDING', None)]

    # Restarting without continuing existing trials should lead to the results
    # [2, 1, 3, 0], because
    # the TERMINATED trial is finished (state = 2),
    # the RUNNING trial is not continued (marked as terminated),
    # the ERROR trial is not continued (remains at 1 and errored)
    # and the PENDING trial is not continued (marked as terminated).

    del tuner
    fail_marker.remove(ignore_errors=True)
    hang_marker.remove(ignore_errors=True)

    tuner = Tuner.restore(
        str(tmpdir / "test_tuner_resume_errored_only"),
        resume_unfinished=False,
        resume_errored=True,
    )
    tuner._local_tuner._run_config.callbacks = None

    results = tuner.fit()
    assert len(results) == 4
    assert len(results.errors) == 0
    assert sorted([r.metrics.get("it", 0) for r in results]) == sorted([2, 1, 3, 0])


def test_tuner_restore_from_cloud(ray_start_2_cpus, tmpdir, clear_memory_filesys):
    """Check that restoring Tuner() objects from cloud storage works"""
    tuner = Tuner(
        lambda config: 1,
        run_config=RunConfig(
            name="exp_dir",
            local_dir=str(tmpdir / "ray_results"),
            sync_config=tune.SyncConfig(upload_dir="memory:///test/restore"),
        ),
    )
    tuner.fit()

    check_path = tmpdir / "check_save"
    download_from_uri("memory:///test/restore", str(check_path))
    remote_contents = os.listdir(check_path / "exp_dir")

    assert "tuner.pkl" in remote_contents
    assert "trainable.pkl" in remote_contents

    prev_cp = _find_newest_experiment_checkpoint(str(check_path / "exp_dir"))
    prev_lstat = os.lstat(prev_cp)

    (tmpdir / "ray_results").remove(ignore_errors=True)

    tuner2 = Tuner.restore("memory:///test/restore/exp_dir")
    results = tuner2.fit()

    assert results[0].metrics["_metric"] == 1
    local_contents = os.listdir(tmpdir / "ray_results" / "exp_dir")
    assert "tuner.pkl" in local_contents
    assert "trainable.pkl" in local_contents

    after_cp = _find_newest_experiment_checkpoint(
        str(tmpdir / "ray_results" / "exp_dir")
    )
    after_lstat = os.lstat(after_cp)

    # Experiment checkpoint was updated
    assert os.path.basename(prev_cp) != os.path.basename(after_cp)
    # Old experiment checkpoint still exists in dir
    assert os.path.basename(prev_cp) in local_contents
    # Contents changed
    assert prev_lstat.st_size != after_lstat.st_size

    # Overwriting should work
    tuner3 = Tuner.restore("memory:///test/restore/exp_dir")
    tuner3.fit()


@pytest.mark.parametrize(
    "upload_uri",
    [None, "memory:///test/test_tuner_restore_latest_available_checkpoint"],
)
def test_tuner_restore_latest_available_checkpoint(
    ray_start_4_cpus, tmpdir, upload_uri, clear_memory_filesys
):
    """Resuming errored trials should pick up from previous state"""
    fail_marker = tmpdir / "fail_marker"
    fail_marker.write_text("", encoding="utf-8")

    tuner = Tuner(
        _train_fn_sometimes_failing,
        tune_config=TuneConfig(
            num_samples=1,
        ),
        run_config=RunConfig(
            name="test_tuner_restore_latest_available_checkpoint",
            local_dir=str(tmpdir),
            sync_config=tune.SyncConfig(upload_dir=upload_uri),
        ),
        param_space={"failing_hanging": (fail_marker, None), "num_epochs": 4},
    )

    results = tuner.fit()

    assert len(results) == 1
    assert len(results.errors) == 1

    [result] = list(results)
    # num_epochs = 4, so it = 4
    assert result.metrics["it"] == 4

    del tuner
    fail_marker.remove(ignore_errors=True)

    shutil.rmtree(os.path.join(result.log_dir, "checkpoint_000003"))
    shutil.rmtree(os.path.join(result.log_dir, "checkpoint_000002"))

    if upload_uri:
        delete_at_uri(upload_uri + "/checkpoint_000003")
        delete_at_uri(upload_uri + "/checkpoint_000002")

    tuner = Tuner.restore(
        str(tmpdir / "test_tuner_restore_latest_available_checkpoint"),
        resume_errored=True,
    )
    results = tuner.fit()
    assert len(results) == 1
    assert len(results.errors) == 0
    [result] = list(results)
    # restored from 2, plus num_epochs = 4, plus one additional epoch
    assert result.metrics["it"] == 7
    assert result.metrics["iterations_since_restore"] == 5


@pytest.mark.parametrize("retry_num", [0, 1])
def test_restore_retry(ray_start_4_cpus, tmpdir, retry_num):
    """Test retrying restore on a trial level."""

    class MockTrainable(Trainable):
        def setup(self, config):
            self.idx = 0
            self.tag_file_path = config["tag_file_path"]
            self._is_restored = False

        def step(self):
            time.sleep(1)
            if self.idx == 0 and self._is_restored:
                raise RuntimeError(
                    "===== Restored trial cannot start from scratch ====="
                )
            elif self.idx == 2 and not self._is_restored:
                raise RuntimeError("===== First run fails at idx=2 =====")
            self.idx += 1
            return {"score": self.idx}

        def save_checkpoint(self, checkpoint_dir):
            path = os.path.join(checkpoint_dir, "checkpoint")
            with open(path, "w") as f:
                f.write(json.dumps({"idx": self.idx}))
            return path

        def load_checkpoint(self, checkpoint_path):
            self._is_restored = True
            if not os.path.exists(self.tag_file_path):
                Path(self.tag_file_path).touch()
                raise RuntimeError("===== Failing first restore =====")
            # The following restore should pass!
            with open(checkpoint_path) as f:
                self.idx = json.loads(f.read())["idx"]

    # Set environment variable just for this test
    with unittest.mock.patch.dict(
        os.environ, {"TUNE_RESTORE_RETRY_NUM": str(retry_num)}
    ):
        tag_file = os.path.join(tmpdir, "tag")
        tuner = Tuner(
            MockTrainable,
            run_config=RunConfig(
                name="tryout_restore",
                stop={"training_iteration": 5},
                local_dir=str(tmpdir),
                failure_config=FailureConfig(max_failures=1),
                checkpoint_config=CheckpointConfig(checkpoint_frequency=1),
            ),
            param_space={"tag_file_path": tag_file},
        )
        results = tuner.fit()
        [result] = list(results)
        if retry_num == 1:
            assert result.metrics["score"] == 5
        else:
            assert result.metrics["score"] == 2


def test_restore_overwrite_trainable(ray_start_4_cpus, tmpdir, caplog):
    """Test validation for trainable compatibility, when re-specifying a trainable
    on restore."""

    def train_func_1(config):
        data = {"data": config["data"]}
        session.report(data, checkpoint=Checkpoint.from_dict(data))
        raise RuntimeError("Failing!")

    tuner = Tuner(
        train_func_1,
        run_config=RunConfig(name="overwrite_trainable", local_dir=str(tmpdir)),
        param_space={"data": 1},
    )
    tuner.fit()

    del tuner

    # Can't overwrite with a different Trainable type
    with pytest.raises(ValueError):
        tuner = Tuner.restore(
            str(tmpdir / "overwrite_trainable"),
            overwrite_trainable="__fake",
            resume_errored=True,
        )

    # Can't overwrite with a different Trainable name
    def train_func_2(config):
        checkpoint = session.get_checkpoint()
        assert checkpoint and checkpoint.to_dict()["data"] == config["data"]

    with pytest.raises(ValueError):
        tuner = Tuner.restore(
            str(tmpdir / "overwrite_trainable"),
            overwrite_trainable=train_func_2,
            resume_errored=True,
        )

    # Can still change trainable code, but logs a warning
    def train_func_1(config):
        checkpoint = session.get_checkpoint()
        assert checkpoint and checkpoint.to_dict()["data"] == config["data"]

    caplog.clear()
    with caplog.at_level(logging.WARNING, logger="ray.tune.impl.tuner_internal"):
        tuner = Tuner.restore(
            str(tmpdir / "overwrite_trainable"),
            overwrite_trainable=train_func_1,
            resume_errored=True,
        )
        assert "The trainable will be overwritten" in caplog.text

    results = tuner.fit()
    assert not results.errors


@pytest.mark.parametrize("use_function_trainable", [True, False])
def test_restore_with_parameters(ray_start_4_cpus, tmp_path, use_function_trainable):
    """Tests Tuner restoration for a `tune.with_parameters` wrapped trainable."""

    def train_func(config, data_str=None, data_obj=None):
        assert data_str is not None and data_obj is not None
        fail_marker = config.pop("fail_marker", None)
        config["failing_hanging"] = (fail_marker, None)
        _train_fn_sometimes_failing(config)

    class FailingTrainable(Trainable):
        def setup(self, config, data_str=None, data_obj=None):
            assert data_str is not None and data_obj is not None
            self.idx = 0
            self.fail_marker = config.get("fail_marker", None)

        def step(self):
            if self.fail_marker and self.fail_marker.exists():
                raise RuntimeError("==== Run is failing ====")
            self.idx += 1
            return {"score": self.idx}

        def save_checkpoint(self, checkpoint_dir):
            return {"idx": self.idx}

        def load_checkpoint(self, checkpoint_dict):
            self.idx = checkpoint_dict["idx"]

    trainable = train_func if use_function_trainable else FailingTrainable

    def create_trainable_with_params():
        data = MockData()
        trainable_with_params = tune.with_parameters(
            trainable, data_str="data", data_obj=data
        )
        return trainable_with_params

    exp_name = "restore_with_params"
    fail_marker = tmp_path / "fail_marker"
    fail_marker.write_text("", encoding="utf-8")

    tuner = Tuner(
        create_trainable_with_params(),
        run_config=RunConfig(
            name=exp_name,
            local_dir=str(tmp_path),
            stop={"training_iteration": 3},
            failure_config=FailureConfig(max_failures=0),
            checkpoint_config=CheckpointConfig(
                checkpoint_frequency=0 if use_function_trainable else 1
            ),
        ),
        param_space={"fail_marker": fail_marker},
    )
    results = tuner.fit()
    assert results.errors

    tuner = Tuner.restore(
        str(tmp_path / exp_name),
        resume_errored=True,
    )
    # Should still be able to access results
    assert len(tuner.get_results().errors) == 1

    # Continuing to fit should fail if we didn't re-specify the trainable
    with pytest.raises(ValueError):
        tuner.fit()

    fail_marker.unlink()
    tuner = Tuner.restore(
        str(tmp_path / exp_name),
        resume_errored=True,
        overwrite_trainable=create_trainable_with_params(),
    )
    results = tuner.fit()
    assert not results.errors


@pytest.mark.parametrize("use_tune_run", [True, False])
def test_tuner_restore_from_moved_experiment_path(
    ray_start_2_cpus, tmp_path, use_tune_run
):
    """Check that restoring a Tuner from a moved experiment directory works."""
    # Create a fail_marker dummy file that causes the first Tune run to fail and
    # the second run to succeed
    fail_marker = tmp_path / "fail_marker"
    fail_marker.write_text("", encoding="utf-8")

    old_local_dir = tmp_path / "ray_results"
    old_exp_name = "exp_dir"

    new_local_dir = tmp_path / "new_ray_results"
    new_exp_name = "new_exp_dir"

    # Initial training run (that errors out in the middle)
    num_to_keep = 2
    tuner = Tuner(
        _train_fn_sometimes_failing,
        tune_config=TuneConfig(
            num_samples=1,
        ),
        run_config=RunConfig(
            name=old_exp_name,
            local_dir=str(old_local_dir),
            checkpoint_config=CheckpointConfig(num_to_keep=num_to_keep),
        ),
        param_space={
            "failing_hanging": (fail_marker, None),
        },
    )
    results = tuner.fit()
    assert len(results.errors) == 1
    training_iteration = results[0].metrics["training_iteration"]
    assert (
        training_iteration == 1
    ), f"Should only have 1 session.report before erroring, got {training_iteration}"

    # Move experiment from `tmp_path/ray_results/exp_dir`
    # to `tmp_path/moved_ray_results/new_exp_dir`, changing both `local_dir` and
    # the experiment `name`
    shutil.move(str(old_local_dir), str(new_local_dir))
    os.rename(str(new_local_dir / old_exp_name), str(new_local_dir / new_exp_name))

    del tuner
    # Remove fail_marker so that the restored Tuner doesn't error again
    fail_marker.unlink()

    # Restore from moved experiment directory location, and launch resumed training
    if use_tune_run:
        analysis = tune.run(
            _train_fn_sometimes_failing,
            name=new_exp_name,
            local_dir=str(new_local_dir),
            resume="AUTO+ERRORED",
        )
        results = ResultGrid(analysis)
    else:
        restore_path = str(new_local_dir / new_exp_name)
        tuner = Tuner.restore(restore_path, resume_errored=True)
        results = tuner.fit()

    assert len(results.errors) == 0
    # Check that we restored iter=1, then made 2 calls to session.report -> iter=3
    training_iteration = results[0].metrics["training_iteration"]
    assert training_iteration == 3, training_iteration

    # Make sure that checkpoints are loaded properly
    assert results[0].checkpoint
    assert len(results[0].best_checkpoints) == num_to_keep
    checkpoint_dirs = [
        path
        for path in os.listdir(results[0].log_dir)
        if path.startswith("checkpoint_")
    ]
    assert sorted(checkpoint_dirs) == ["checkpoint_000001", "checkpoint_000002"]

    # Make sure that we did not create a logdir in the old location
    assert not old_local_dir.exists()


def test_tuner_restore_from_moved_cloud_uri(
    ray_start_2_cpus, tmp_path, clear_memory_filesys
):
    """Test that restoring an experiment that was moved to a new remote URI
    resumes and continues saving new results at that URI."""

    def failing_fn(config):
        data = {"score": 1}
        session.report(data, checkpoint=Checkpoint.from_dict(data))
        raise RuntimeError("Failing!")

    tuner = Tuner(
        failing_fn,
        run_config=RunConfig(
            name="exp_dir",
            local_dir=str(tmp_path / "ray_results"),
            sync_config=tune.SyncConfig(upload_dir="memory:///original"),
        ),
        tune_config=TuneConfig(trial_dirname_creator=lambda _: "test"),
    )
    tuner.fit()

    # mv memory:///original/exp_dir memory:///moved/new_exp_dir
    download_from_uri(
        "memory:///original/exp_dir", str(tmp_path / "moved" / "new_exp_dir")
    )
    delete_at_uri("memory:///original")
    upload_to_uri(str(tmp_path / "moved"), "memory:///moved")

    tuner = Tuner.restore("memory:///moved/new_exp_dir", resume_errored=True)
    # Just for the test, since we're using `memory://` to mock a remote filesystem,
    # the checkpoint needs to be copied to the new local directory.
    # This is because the trainable actor uploads its checkpoints to a
    # different `memory://` filesystem than the driver and is not
    # downloaded along with the other parts of the experiment dir.
    # NOTE: A new local directory is used since the experiment name got modified.
    shutil.move(
        tmp_path / "ray_results/exp_dir/test/checkpoint_000000",
        tmp_path / "ray_results/new_exp_dir/test/checkpoint_000000",
    )
    results = tuner.fit()

    assert list_at_uri("memory:///") == ["moved"]
    num_experiment_checkpoints = len(
        [
            path
            for path in list_at_uri("memory:///moved/new_exp_dir")
            if path.startswith("experiment_state")
        ]
    )
    assert num_experiment_checkpoints == 2

    num_trial_checkpoints = len(
        [
            path
            for path in os.listdir(results[0].log_dir)
            if path.startswith("checkpoint_")
        ]
    )
    assert num_trial_checkpoints == 2


def test_restore_from_relative_path(ray_start_4_cpus, chdir_tmpdir):
    tuner = Tuner(
        lambda config: session.report({"score": 1}),
        run_config=RunConfig(local_dir="relative_dir", name="exp_name"),
    )
    tuner.fit()

    tuner = Tuner.restore("relative_dir/exp_name")
    results = tuner.fit()
    assert not results.errors
    assert results[0].metrics["score"] == 1


def test_custom_searcher_and_scheduler_restore(ray_start_2_cpus, tmpdir):
    """Check that a restored Tune experiment uses the original searcher/scheduler."""
    fail_marker = tmpdir / "fail_marker"
    fail_marker.write_text("", encoding="utf-8")

    class MockSearcher(OptunaSearch):
        def on_trial_result(self, trial_id: str, result: dict):
            super().on_trial_result(trial_id, result)
            if not hasattr(self, "_test_result_counter"):
                self._test_result_counter = 0
            self._test_result_counter += 1

    class MockScheduler(ASHAScheduler):
        def on_trial_result(self, runner, trial, result):
            decision = super().on_trial_result(runner, trial, result)
            if not hasattr(self, "_test_result_counter"):
                self._test_result_counter = 0
            self._test_result_counter += 1
            return decision

    tuner = Tuner(
        _train_fn_sometimes_failing,
        run_config=RunConfig(local_dir=str(tmpdir), name="exp_name"),
        tune_config=TuneConfig(
            search_alg=MockSearcher(),
            scheduler=MockScheduler(),
            metric="it",
            mode="max",
        ),
        param_space={"a": tune.uniform(0, 1), "failing_hanging": (fail_marker, None)},
    )
    tuner.fit()

    del tuner
    fail_marker.remove(ignore_errors=True)

    tuner = Tuner.restore(str(tmpdir / "exp_name"), resume_errored=True)
    tuner.fit()
    searcher = tuner._local_tuner._tune_config.search_alg
    scheduler = tuner._local_tuner._tune_config.scheduler
    assert isinstance(searcher, MockSearcher)
    assert isinstance(scheduler, MockScheduler)
    # Searcher state should get loaded correctly
    # Total of 3 reported results (1 from before failure, 2 after restore)
    assert searcher._test_result_counter == 3
    # Make sure that the restored scheduler is at least used
    assert (
        hasattr(scheduler, "_test_result_counter")
        and scheduler._test_result_counter > 0
    )


@pytest.mark.parametrize("use_air_trainer", [True, False])
def test_checkpoints_saved_after_resume(ray_start_2_cpus, tmp_path, use_air_trainer):
    """Checkpoints saved after experiment restore should pick up at the correct
    iteration and should not overwrite the checkpoints from the original run.
    Old checkpoints should still be deleted if the total number of checkpoints
    (old + new) exceeds `num_to_keep`.

    In this test, `num_to_keep=4`:
    - Initial run saves checkpoint_000000 and checkpoint_000001
    - Restored run saves checkpoint_000002, checkpoint_000003, and checkpoint_000004
    - Checkpoint 000000 should be deleted.
    """

    def get_checkpoints(experiment_dir):
        checkpoint_dirs = [
            path
            for path in os.listdir(experiment_dir)
            if path.startswith("checkpoint_")
        ]
        sorted_checkpoint_dirs = sorted(checkpoint_dirs)
        checkpoints = [
            Checkpoint.from_directory(os.path.join(experiment_dir, d))
            for d in sorted_checkpoint_dirs
        ]
        return sorted_checkpoint_dirs, checkpoints

    fail_marker = tmp_path / "fail_marker"
    fail_marker.write_text("", encoding="utf-8")

    trainable = (
        DataParallelTrainer(
            _train_fn_sometimes_failing, scaling_config=ScalingConfig(num_workers=1)
        )
        if use_air_trainer
        else _train_fn_sometimes_failing
    )
    param_space = {
        "failing_hanging": (fail_marker, None),
        "num_epochs": 2,
    }
    if use_air_trainer:
        param_space = {"train_loop_config": param_space}

    num_to_keep = 4
    tuner = Tuner(
        trainable,
        tune_config=TuneConfig(num_samples=1),
        run_config=RunConfig(
            name="exp_name",
            local_dir=str(tmp_path),
            checkpoint_config=CheckpointConfig(num_to_keep=num_to_keep),
        ),
        param_space=param_space,
    )
    results = tuner.fit()
    training_iteration = results[0].metrics["training_iteration"]
    assert (
        training_iteration == 2
    ), f"Should be at 2 iters before erroring, got {training_iteration}"

    # Initial run saves the first 2 checkpoints
    checkpoint_dirs, checkpoints = get_checkpoints(results[0].log_dir)
    assert checkpoint_dirs == ["checkpoint_000000", "checkpoint_000001"]
    assert [ckpt.to_dict()["it"] for ckpt in checkpoints] == [1, 2]

    fail_marker.unlink()
    tuner = Tuner.restore(str(tmp_path / "exp_name"), resume_errored=True)
    results = tuner.fit()

    assert len(results.errors) == 0
    training_iteration = results[0].metrics["training_iteration"]
    # Restored at it=2, reported 3 more times -> should have it=5
    assert training_iteration == 5

    # Restored run saves the 3 more checkpoints, and first checkpoint should be deleted
    checkpoint_dirs, checkpoints = get_checkpoints(results[0].log_dir)
    assert checkpoint_dirs == [f"checkpoint_00000{i}" for i in range(1, 5)]
    assert [ckpt.to_dict()["it"] for ckpt in checkpoints] == [2, 3, 4, 5]


if __name__ == "__main__":
    import sys

    sys.exit(pytest.main(["-v", __file__]))



import logging
import os
from typing import Optional

import ray
from ray.serve._private.constants import DEBUG_LOG_ENV_VAR, SERVE_LOGGER_NAME

COMPONENT_LOG_FMT = "%(levelname)s %(asctime)s {component_name} {component_id} %(filename)s:%(lineno)d - %(message)s"  # noqa:E501
LOG_FILE_FMT = "{component_name}_{component_id}.log"


def access_log_msg(*, method: str, route: str, status: str, latency_ms: float):
    """Returns a formatted message for an HTTP or ServeHandle access log."""
    return f"{method.upper()} {route} {status.upper()} {latency_ms:.1f}ms"


def configure_component_logger(
    *,
    component_name: str,
    component_id: str,
    component_type: Optional[str] = None,
    log_level: int = logging.INFO,
    log_to_stream: bool = True,
    log_to_file: bool = True,
    max_bytes: Optional[int] = None,
    backup_count: Optional[int] = None,
):
    """Returns a logger to be used by a Serve component.

    The logger will log using a standard format to make components identifiable
    using the provided name and unique ID for this instance (e.g., replica ID).

    This logger will *not* propagate its log messages to the parent logger(s).
    """
    logger = logging.getLogger(SERVE_LOGGER_NAME)
    logger.propagate = False
    logger.setLevel(log_level)
    if os.environ.get(DEBUG_LOG_ENV_VAR, "0") != "0":
        logger.setLevel(logging.DEBUG)

    formatter = logging.Formatter(
        COMPONENT_LOG_FMT.format(
            component_name=component_name, component_id=component_id
        )
    )
    if log_to_stream:
        stream_handler = logging.StreamHandler()
        stream_handler.setFormatter(formatter)
        logger.addHandler(stream_handler)

    if log_to_file:
        logs_dir = os.path.join(
            ray._private.worker._global_node.get_logs_dir_path(), "serve"
        )
        os.makedirs(logs_dir, exist_ok=True)
        if max_bytes is None:
            max_bytes = ray._private.worker._global_node.max_bytes
        if backup_count is None:
            backup_count = ray._private.worker._global_node.backup_count
        if component_type is not None:
            component_name = f"{component_type}_{component_name}"
        log_file_name = LOG_FILE_FMT.format(
            component_name=component_name, component_id=component_id
        )
        file_handler = logging.handlers.RotatingFileHandler(
            os.path.join(logs_dir, log_file_name),
            maxBytes=max_bytes,
            backupCount=backup_count,
        )
        file_handler.setFormatter(formatter)
        logger.addHandler(file_handler)


class LoggingContext:
    """
    Context manager to manage logging behaviors within a particular block, such as:
    1) Overriding logging level

    Source (python3 official documentation)
    https://docs.python.org/3/howto/logging-cookbook.html#using-a-context-manager-for-selective-logging # noqa: E501
    """

    def __init__(self, logger, level=None):
        self.logger = logger
        self.level = level

    def __enter__(self):
        if self.level is not None:
            self.old_level = self.logger.level
            self.logger.setLevel(self.level)

    def __exit__(self, et, ev, tb):
        if self.level is not None:
            self.logger.setLevel(self.old_level)
